"""Menu TUI renderable."""
# dem/cli/tui/renderable/menu.py

from dem.cli.tui.printable_tool_image import PrintableToolImage
from rich import live, table, align, panel, box
from rich.console import RenderableType

from readchar import readkey, key

class BaseMenu(table.Table):
    """ Base class for the menus.
    
        Class attributes:
            cursor_on -- text that represents the cursor
            cursor_off -- the cursor is not in this cell
    """
    cursor_on = "* "
    cursor_off = "  "

    def __init__(self, box: box.Box | None = box.HEAVY_HEAD) -> None:
        """ Construct the BaseMenu based on the rich's Table class.
        
            Args:
                box -- how to draw the table's edges (None: no outline)
        """
        super().__init__(box=box)
        self.cursor_pos = 0

    def remove_cursor(self, column: int = 0, row: int = 0) -> None:
        """ Remove the cursor from the given cell.
        
            Args:
                column -- the column index
                row -- the row index
        """
        cell = self.columns[column]._cells[row].replace(self.cursor_on, self.cursor_off, 1)
        self.columns[column]._cells[row] = cell

    def add_cursor(self, column: int = 0, row: int = 0) -> None:
        """ Add the cursor to the given cell.
        
            Args:
                column -- the column index
                row -- the row index
        """
        cell = self.columns[column]._cells[row].replace(self.cursor_off, self.cursor_on, 1)
        self.columns[column]._cells[row] = cell

    def set_title(self, title: str) -> None:
        """ Set the text above the menu.
        
            Args:
                title -- the text to set
        """
        self.title = title + "\n"

class VerticalMenu(BaseMenu):
    """ Menu with vertical navigation.
    
        Shouldn't be instantiated directly.
        
        Class attributes:
            CURSOR_UP -- move cursor upwards
            CURSOR_DOWN -- move cursor downwards
    """
    (
        CURSOR_UP,
        CURSOR_DOWN
    ) = range(2)

    def __init__(self) -> None:
        """ Construct a VerticalMenu based on the BaseMenu."""
        super().__init__()

    def move_cursor(self, cursor_direction: int) -> None:
        """ Move the cursor in the given direction.
        
            Args:
                cursor_direction -- which direction to move
        """
        self.remove_cursor()

        if (cursor_direction == self.CURSOR_UP):
            if self.cursor_pos == 0:
                self.cursor_pos = self.row_count - 1
            else:
                self.cursor_pos -= 1
        else:
            if self.cursor_pos == self.row_count - 1:
                self.cursor_pos = 0
            else:
                self.cursor_pos += 1

        self.add_cursor()
 
    def remove_cursor(self) -> None:
        """ Remove the cursor."""
        super().remove_cursor(row=self.cursor_pos)

    def add_cursor(self) -> None:
        """ Add the cursor."""
        super().add_cursor(row=self.cursor_pos)

    def handle_user_input(self, input: str) -> None:
        """ Handle user input.
        
            Args:
                input -- the user input (handle up/down arrows, j and k)
        """
        match input:
            case key.UP | 'k':
                self.move_cursor(self.CURSOR_UP)
            case key.DOWN | 'j':
                self.move_cursor(self.CURSOR_DOWN)

class HorizontalMenu(BaseMenu):
    """ Menu with horizontal navigation.
    
        Shouldn't be instantiated directly.
        
        Class attributes:
            CURSOR_LEFT -- move cursor left
            CURSOR_RIGHT -- move cursor right
    """
    (
        CURSOR_LEFT,
        CURSOR_RIGHT
    ) = range(2)

    def __init__(self) -> None:
        """ Construct a HorizontalMenu based on the BaseMenu."""
        super().__init__(box=None)

    def move_cursor(self, cursor_direction: int) -> None:
        """ Move the cursor in the given direction.
        
            Args:
                cursor_direction -- which direction to move
        """
        self.remove_cursor()

        if (cursor_direction == self.CURSOR_LEFT):
            if self.cursor_pos == 0:
                self.cursor_pos = len(self.columns) - 1
            else:
                self.cursor_pos -= 1
        else:
            if self.cursor_pos == len(self.columns) - 1:
                self.cursor_pos = 0
            else:
                self.cursor_pos += 1

        self.add_cursor()

    def remove_cursor(self) -> None:
        """ Remove the cursor."""
        super().remove_cursor(column=self.cursor_pos)

    def add_cursor(self) -> None:
        """ Add the cursor."""
        super().add_cursor(column=self.cursor_pos)

    def handle_user_input(self, input:str) -> None:
        """ Handle user input.
        
            Args:
                input -- the user input (handle left/right arrows, h and l)
        """
        match input:
            case key.LEFT | 'h':
                self.move_cursor(self.CURSOR_LEFT)
            case key.RIGHT | 'l':
                self.move_cursor(self.CURSOR_RIGHT)

class CancelSaveMenu(HorizontalMenu):
    """ Horizontal menu with two items: cancel and save.
    
        Class attributes:
            menu_items -- the menu items with rich text formatting
    """
    menu_items = ("[underline]cancel[/]", "[underline]save[/]")
    
    def __init__(self) -> None:
        """ Construct a HorizontalMenu with 2 columns and 1 row."""
        super().__init__()
        self.add_row(self.cursor_off + self.menu_items[0], 
                     self.cursor_off + self.menu_items[1])
        self.is_selected = False

    def handle_user_input(self, input: str) -> None:
        """ Handle user input or pass to parent.
        
            Args:
                input -- the user input (handle enter)
        """
        if input == key.ENTER:
            self.is_selected = True
        else:
            super().handle_user_input(input)

    def get_selection(self) -> str:
        """ Get the selected item.
        
            Return with the selected item.
        """
        return self.columns[self.cursor_pos]._cells[0]

class ToolImageMenu(VerticalMenu):
    def __init__(self, printable_tool_images: list[PrintableToolImage],
                 already_selected_tool_images: list[str]) -> None:
        super().__init__()
        self.title = "Select the tool images for the Development Environment:"
        self.add_column("Tool images", no_wrap=True)
        self.add_column("Availability", no_wrap=True)
        self.tool_image_selection = []

        for index, tool_image in enumerate(printable_tool_images):
            row_content = []
            if (index == 0):
                # Set the cursor indicator for the first element.
                row_content = "* " + tool_image.name, tool_image.status
            else:
                row_content = "  " + tool_image.name, tool_image.status

            if tool_image.name in already_selected_tool_images:
                self.tool_image_selection.append(tool_image.name)
                row_content = f"{row_content[0][:2]}[green]{row_content[0][2:]}[/]", str(row_content[1])
            
            self.add_row(*row_content)
    
    def get_table_width(self) -> int:
        return sum(self._measure_column)

    def handle_user_input(self, input: str) -> None:
        if input == key.ENTER or input == key.SPACE:
            selected_cell = self.columns[0]._cells[self.cursor_pos]

            if "[green]" in selected_cell:
                selected_cell = selected_cell.replace("[green]", "")
                selected_cell = selected_cell.replace("[/]", "")
            else:
                selected_cell = f"{selected_cell[:2]}[green]{selected_cell[2:]}[/]"

            self.columns[0]._cells[self.cursor_pos] = selected_cell
        else:
            super().handle_user_input(input)

    def get_selected_tool_images(self) -> list[str]:
        for cell in self.columns[0]._cells:
            tool_image_name = cell[2:].replace("[/]", "").replace("[green]", "")
            if ("[green]" in cell) and (tool_image_name not in self.tool_image_selection):
                self.tool_image_selection.append(tool_image_name)
            elif ("[green]" not in cell) and (tool_image_name in self.tool_image_selection):
                self.tool_image_selection.remove(tool_image_name)
        
        return self.tool_image_selection

class SelectMenu(VerticalMenu):
    def __init__(self, selection: list[str]) -> None:
        super().__init__()
        self.show_edge = False
        self.show_lines = False

        for index, element in enumerate(selection):
            if (index == 0):
                # Set the cursor indicator for the first element.
                self.add_row("* " + element)
            else:
                self.add_row("  " + element)

        self.alignment = align.Align(self, align="center", vertical="middle")

    def wait_for_user(self):
        with live.Live(self.alignment, refresh_per_second=8, screen=True):
            while True:
                match readkey():
                    case key.UP | 'k':
                        self.move_cursor(self.CURSOR_UP)
                    case key.DOWN | 'j':
                        self.move_cursor(self.CURSOR_DOWN)
                    case key.ENTER:
                        break
            
    def get_selected(self) -> str:
        return self.columns[0]._cells[self.cursor_pos][2:]
    
    def set_title(self, title: str) -> None:
        self.width = len(title)
        super().set_title(title)

class DevEnvStatusPanel(table.Table):
    def __init__(self, selected_tool_images: list[str], height: int, width: int) -> None:
        super().__init__(title="Dev Env Settings")
        self.add_column("Selected Tool Images", no_wrap=True)

        if height > 10:
            height = 10

        for tool_image in selected_tool_images:
            self.add_row(tool_image)

        if len(selected_tool_images) < height:
            for _ in range(height - len(selected_tool_images)):
                self.add_row(" " * width)


"""Implementation of the assign command."""
# dem/cli/command/assign_cmd.py

import os
from dem.core.platform import Platform
from dem.core.dev_env import DevEnv
from dem.cli.console import stderr, stdout

def execute(platform: Platform, dev_env_name: str, project_path: str) -> None:
    """Assign the given Development Environment to the given project.

        Args:
            platform -- the platform
            dev_env_name -- the name of the Development Environment to assign
            project_path -- the path to the project to assign the Development Environment to
    """


    if not os.path.isdir(project_path):
        stderr.print(f"[red]Error: The {project_path} path does not exist.[/]")
        return

    dev_env_to_assign: DevEnv | None = platform.get_dev_env_by_name(dev_env_name)

    if dev_env_to_assign is None:
        stderr.print(f"[red]Error: The {dev_env_name} Development Environment does not exist.[/]")
    else:
        platform.assign_dev_env(dev_env_to_assign, project_path)
        stdout.print(f"\n[green]Successfully assigned the {dev_env_name} Dev Env to the project at {project_path}![/]")


"""CLI command implementation for adding a host."""

from dem.core.platform import Platform
from dem.cli.console import stdout
import typer

def execute(platform: Platform, name: str, address: str) -> None:
    """ Add a new host.
    
        Args:
            name -- name of the host
            address -- IP or hostname of the host
    """
    existing_host: dict | None = next((host for host in platform.hosts.list_host_configs() if host["name"] == name), 
                                      None)
    if existing_host:
        typer.confirm(f"Host with name {name} already exists. Do you want to overwrite it?", 
                      abort=True)
        platform.hosts.delete_host(existing_host)

    platform.hosts.add_host({"name": name, "address": address})
    stdout.print("[green]Host added successfully![/]")


"""This module represents a Development Environment."""
# dem/core/dev_env.py

from dem.core.tool_images import ToolImage, ToolImages
from enum import Enum
import json, os

class DevEnv():
    """ A Development Environment. """

    class Status(Enum):
        """ The status of an installed Development Environment. """
        OK = 0
        REINSTALL_NEEDED = 1
        UNAVAILABLE_IMAGE = 2

    def __init__(self, descriptor: dict | None = None, descriptor_path: str | None = None) -> None:
        """ Init the DevEnv class. 
        
            A new instance can be created:
            - from a Dev Env descriptor
            - from a descriptor avaialable at the given path.

            Only one of the arguments can be used at a time.
        
            Args:
                descriptor -- the description of the Development Environment from the dev_env.json 
                              file
                descriptor_path -- the path of the descriptor file

            Exceptions:
                ValueError -- if more than one of the arguments is not None
        """

        # Only one of the arguments can be not None
        if sum(arg is not None for arg in [descriptor, descriptor_path]) > 1:
            raise ValueError("Only one of the arguments can be not None.")

        if descriptor_path:
            if not os.path.exists(descriptor_path):
                raise FileNotFoundError("dev_env_descriptor.json doesn't exist.")
            with open(descriptor_path, "r") as file:
                descriptor = json.load(file)

        self.name: str = descriptor["name"]
        self.tool_image_descriptors: list[dict[str, str]] = descriptor["tools"]
        self.tool_images: list[ToolImage] = []
        descriptor_installed = descriptor.get("installed", "False")
        if "True" == descriptor_installed:
            self.is_installed = True
        else:
            self.is_installed = False

    def assign_tool_image_instances(self, tool_images: ToolImages) -> None:
        """ Assign the Tool Images to the Development Environment.
        
            After creating a DevEnv instance, the Tool Images are not yet assigned to it. This 
            method must be called if the Tool Images are needed.
            
            Args:
                tool_images -- the Tool Images to assign
                
                Exceptions:
                    ToolImageError -- if the Tool Image name is invalid
        """
        for tool_descriptor in self.tool_image_descriptors:
            tool_image_name = tool_descriptor["image_name"] + ':' + tool_descriptor["image_version"]
            tool_image = tool_images.all_tool_images.get(tool_image_name, ToolImage(tool_image_name))
            self.tool_images.append(tool_image)

    def get_tool_image_status(self) -> Status:
        """ Get the status of the Tool Images.

            This method checks the availability of the assigned Tool Images. 
            If at least one of the Tool Images is unkonwn: NOT_AVAILABLE. 
            If at least one of the Tool Images is only available in the registry: REINSTALL_NEEDED. 
            If all the Tool Images are available: OK.

            Returns:
                Status -- the status of the Dev Env
        """
        for tool_image in self.tool_images:
            if tool_image.availability == ToolImage.NOT_AVAILABLE:
                return self.Status.UNAVAILABLE_IMAGE
            elif tool_image.availability == ToolImage.REGISTRY_ONLY:
                return self.Status.REINSTALL_NEEDED
        return self.Status.OK

    def get_deserialized(self, omit_is_installed: bool = False) -> dict[str, str]:
        """ Create the deserialized json. 
        
            Return the Dev Env as a dict.
        """
        dev_env_json_deserialized: dict = {
            "name": self.name,
            "tools": self.tool_image_descriptors
        }
        
        if omit_is_installed is False:
            if self.is_installed:
                dev_env_json_deserialized["installed"] = "True"
            else:
                dev_env_json_deserialized["installed"] = "False"

        return dev_env_json_deserialized

    def export(self, path: str) -> None:
        """ Export the Dev Env to a file.
        
            Args:
                path -- the path of the file to export the Dev Env to
        """
        with open(path, "w") as file:
            json.dump(self.get_deserialized(True), file, indent=4)


"""The clone command"""
# dem/cli/command/clone_cmd.py

import typer
from dem.core.platform import Platform
from dem.core.dev_env import DevEnv
from dem.core.exceptions import PlatformError
from dem.cli.console import stdout, stderr

def handle_existing_local_dev_env(platform: Platform, local_dev_env: DevEnv) -> None:
    """ Handle the case when the Dev Env already exists locally.

        Args:
            platform -- the platform
            local_dev_env -- the existing Dev Env
    """
    stdout.print("[yellow]The Dev Env already exists.[/]")
    typer.confirm("Continue with overwrite?", abort=True)
    
    if local_dev_env.is_installed:
        typer.confirm("The Dev Env to overwrite is installed. Do you want to uninstall it?", 
                      abort=True)
        try:
            platform.uninstall_dev_env(local_dev_env)
        except PlatformError as e:
            stderr.print(f"[red]{str(e)}[/]")
            raise typer.Abort()

    platform.local_dev_envs.remove(local_dev_env)

def execute(platform: Platform, dev_env_name: str) -> None:
    """ Copy the Dev Env's descriptor from the catalog to the local descriptor storage.

        If the Dev Env already exists locally, the user will be asked to confirm the overwrite.

        Args:
            platform -- the platform
            dev_env_name -- name of the Dev Env to clone
    """


    catalog_dev_env: DevEnv | None = None

    if not platform.dev_env_catalogs.catalogs:
        stderr.print("[red]Error: No Development Environment Catalogs are available to clone from![/]")
        return

    for catalog in platform.dev_env_catalogs.catalogs:
        catalog.request_dev_envs()
        catalog_dev_env = catalog.get_dev_env_by_name(dev_env_name)
        if catalog_dev_env:
            break
    else:
        stderr.print("[red]Error: The input Development Environment is not available.[/]")
        return

    local_dev_env: DevEnv | None = platform.get_dev_env_by_name(dev_env_name)
    if local_dev_env:
        handle_existing_local_dev_env(platform, local_dev_env)

    platform.local_dev_envs.append(catalog_dev_env)
    platform.flush_descriptors()

    stdout.print("[green]The Dev Env successfully cloned.[/]")


"""All core module must inherit from the Core class."""
# dem/core/core.py

from dem.core.user_output import UserOutput, NoUserOutput
from dem.core.data_management import ConfigFile

class Core():
    """ Base class for all core classes.
    
        Class attributes:
            user_output -- interface to the UI (must be a descendant of the UserOutput class)
    """
    user_output: UserOutput = NoUserOutput()
    config_file: ConfigFile = ConfigFile()

    """ Set the user output class for all core descendant core classes.
    
        Args:
            user_output -- interface to the UI (must be a descendant of the UserOutput class)
    """
    @classmethod
    def set_user_output(cls, user_output: UserOutput) -> None:
        cls.user_output = user_output


"""Collection of dem specific exceptions."""
# dem/core/exceptions.py

class DataStorageError(Exception):
    """Raised when valid data can't be loaded from file."""

    base_message = "Invalid file: "

    def __init__(self, message: str = "") -> None:
        super().__init__(self.base_message + message)

class RegistryError(Exception):
    """Raised when the communication with registry fails."""

    base_message = "Registry error: "

    def __init__(self, message: str) -> None:
        super().__init__(self.base_message + message)

class ToolImageError(Exception):
    """Raised when there is a problem with the tool image."""
    pass

class ContainerEngineError(Exception):
    """Raised when there is a problem with the container engine."""

    base_message = "Container engine error: "

    def __init__(self, message: str) -> None:
        super().__init__(self.base_message + message)

class InternalError(Exception):
    """Raised when an object is used incorrectly."""
    pass

class PlatformError(Exception):
    """Raised when there is a problem with the platform."""

    base_message = "Platform error: "

    def __init__(self, message: str) -> None:
        super().__init__(self.base_message + message)

class CatalogError(Exception):
    """Raised when there is a problem with the catalog."""

    base_message = "Catalog error: "

    def __init__(self, message: str) -> None:
        super().__init__(self.base_message + message)


"""modify CLI command implementation."""
# dem/cli/command/modify_cmd

import copy, typer
from dem.core.dev_env import DevEnv, DevEnv
from dem.core.tool_images import ToolImage
from dem.core.platform import Platform
from dem.core.exceptions import PlatformError
from dem.cli.console import stderr, stdout
from dem.cli.tui.renderable.menu import SelectMenu
from dem.cli.tui.window.dev_env_settings_window import DevEnvSettingsWindow
from dem.cli.tui.printable_tool_image import PrintableToolImage, convert_to_printable_tool_images

def get_confirm_from_user() -> str:
    """ Get the confirmation from the user.
        
        Returns:
            the confirmation option
    """
    confirm_menu_items = ["confirm", "save as", "cancel"]
    select_menu = SelectMenu(confirm_menu_items)
    select_menu.set_title("Are you sure to overwrite the Development Environment?")
    select_menu.wait_for_user()
    return select_menu.get_selected()

def handle_user_confirm(confirmation: str, dev_env_local: DevEnv, platform: Platform) -> None:
    """ Handle the user confirmation.
    
        Args:
            confirmation -- the confirmation option
            dev_env_local -- the local Development Environment
            platform -- the platform

        Exceptions:
            typer.Abort -- when the user tries to save as the Dev Env with a taken name or cancels 
                           the operation
    """
    if confirmation == "cancel":
        raise typer.Abort()

    if confirmation == "save as":
        new_dev_env = copy.deepcopy(dev_env_local)
        new_dev_env.name = typer.prompt("Name of the new Development Environment")
        
        check_for_new_dev_env = platform.get_dev_env_by_name(new_dev_env.name)

        if check_for_new_dev_env is None:
            platform.local_dev_envs.append(new_dev_env)
        else:
            stderr.print("[red]The Development Environment already exist.")
            raise typer.Abort()

    platform.flush_descriptors()

def get_already_selected_tool_images(dev_env: DevEnv) -> set[str]:
    """ Get the already selected Tool Images.
    
        Args:
            dev_env -- the Development Environment
            
        Returns:
            the already selected Tool Images
    """
    already_selected_tool_images = []
    for tool in dev_env.tool_image_descriptors:
        already_selected_tool_images.append(tool["image_name"] + ":" + tool["image_version"])

    return already_selected_tool_images

def remove_missing_tool_images(all_tool_images: dict[str, ToolImage], 
                               already_selected_tool_images: list[str]) -> None:
    """ Remove the missing Tool Images from the Development Environment.
        
            Args:
                all_tool_images -- all available Tool Images
                already_selected_tool_images -- the already selected Tool Images
            
            Exceptions:
                typer.Abort -- if the user doesn't want the removal of the missing Tool Images
    """
    tool_images_are_missing = False

    for already_selected_tool_image in already_selected_tool_images:
        try:
            all_tool_images[already_selected_tool_image]
        except KeyError:
            stderr.print(f"[red]The {already_selected_tool_image} is not available anymore.[/]")
            already_selected_tool_images.remove(already_selected_tool_image)
            tool_images_are_missing = True
    
    if tool_images_are_missing:
        typer.confirm("By continuing, the missing tool images will be removed from the Development Environment.", 
                    abort=True)

def open_dev_env_settings_panel(already_selected_tool_images: list[str], 
                                printable_tool_images: list[PrintableToolImage]) -> list[str]:
    """ Open the Development Environment settings panel.
    
        Args:
            already_selected_tool_images -- the already selected Tool Images
            printable_tool_images -- the printable Tool Images
        
        Returns:
            the selected Tool Images

        Exceptions:
            typer.Abort -- if the user cancels the operation
    """
    dev_env_settings_panel = DevEnvSettingsWindow(printable_tool_images, already_selected_tool_images)
    dev_env_settings_panel.wait_for_user()

    if "cancel" in dev_env_settings_panel.cancel_save_menu.get_selection():
        raise typer.Abort()

    return dev_env_settings_panel.tool_image_menu.get_selected_tool_images()

def update_dev_env(dev_env: DevEnv, selected_tool_images: list[str]) -> None:
    """ Update the Development Environment.
    
        Args:
            dev_env -- the Development Environment
            selected_tool_images -- the selected Tool Images
    """
    dev_env.tool_image_descriptors = []

    for tool_image in selected_tool_images:
        if "/" in tool_image:
            registry, image = tool_image.split("/")
            image_name = registry + '/' + image.split(":")[0]
        else:
            image = tool_image
            image_name = image.split(":")[0]

        dev_env.tool_image_descriptors.append({
            "image_name": image_name,
            "image_version": image.split(":")[1]
        })

def modify_with_tui(platform: Platform, dev_env: DevEnv) -> None:
    """ Modify the Dev Env with the TUI.
    
        Args:
            platform -- the platform
            dev_env -- the Development Environment
            
        Exceptions:
            typer.Abort -- if the user cancels the operation
    """
    already_selected_tool_images = get_already_selected_tool_images(dev_env)
    remove_missing_tool_images(platform.tool_images.all_tool_images, already_selected_tool_images)
    printable_tool_images = convert_to_printable_tool_images(platform.tool_images.all_tool_images)
    selected_tool_images = open_dev_env_settings_panel(already_selected_tool_images, 
                                                       printable_tool_images)
    update_dev_env(dev_env, selected_tool_images)
    confirmation = get_confirm_from_user()
    handle_user_confirm(confirmation, dev_env, platform)

def execute(platform: Platform, dev_env_name: str) -> None:
    """ Modify the Development Environment.
    
        Args:
            platform -- the platform
            dev_env_name -- the name of the Development Environment
            
        Exceptions:
            typer.Abort -- if the user cancels the operation
    """

    platform.assign_tool_image_instances_to_all_dev_envs()

    dev_env = platform.get_dev_env_by_name(dev_env_name)

    if dev_env is None:
        stderr.print("[red]The Development Environment doesn't exist.")
        return
    elif dev_env.is_installed is True:
        stdout.print("[yellow]The Development Environment is installed, so it can't be modified.[/]")
        typer.confirm("Do you want to uninstall it first?", abort=True)
        try:
            platform.uninstall_dev_env(dev_env)
        except PlatformError as e:
            stderr.print(f"[red]{str(e)}[/]")
            return

    modify_with_tui(platform, dev_env)
    stdout.print("[green]The Development Environment has been modified successfully![/]")


"""A class representing a tool image with its printable availability."""
# dem/cli/tui/printable_tool_image.py

from dem.core.tool_images import ToolImage

tool_image_statuses = {
    ToolImage.LOCAL_ONLY: "local",
    ToolImage.REGISTRY_ONLY: "registry",
    ToolImage.LOCAL_AND_REGISTRY: "local and registry"
}

class PrintableToolImage():
    def __init__ (self, tool_image: ToolImage):
        self.name = tool_image.name
        self.status = tool_image_statuses[tool_image.availability]

def convert_to_printable_tool_images(all_tool_images: dict[str, ToolImage]) -> list[PrintableToolImage]:
    printable_tool_images = []
    # sorted will return a list of tuples, so we can iterate over the items
    for tool_image in sorted(all_tool_images.items()):
        # tool_image[0] is the name of the tool image and tool_image[1] is the ToolImage instance
        printable_tool_images.append(PrintableToolImage(tool_image[1]))

    return printable_tool_images


"""CLI command implementation for deleting a registry."""
# dem/cli/command/del_reg_cmd.py

from dem.core.platform import Platform
from dem.cli.console import stdout, stderr

def execute(platform: Platform, registry_name: str) -> None:
    """ Delete the registry.
        Args:
            registry_name -- name of the registry to delete
    """
    for registry in platform.registries.list_registry_configs():
        if registry["name"] == registry_name:
            platform.registries.delete_registry(registry)
            stdout.print("[green]The input registry has been successfully deleted.")
            break
    else:
        stderr.print("[red]Error: The input registry name doesn't exist![/]")


"""run CLI command implementation."""
# dem/cli/command/run_cmd.py

from dem.core.dev_env import DevEnv
from dem.core.platform import Platform
from dem.cli.console import stdout, stderr
import typer

def handle_missing_tool_images(missing_tool_images: set[str], dev_env_local: DevEnv,
                               platform: Platform) -> None:
    """ Report an error to the user that some images are not available. Ask them if the DEM should 
        try to fix the Dev Env: abort if no, pull the missing tool images if yes.
        
        Args:
            missing_tool_images -- the missing tool images
            dev_env_local -- local Dev Env
            platform -- provides the interface to pull the missing images
    """
    stderr.print("[red]Error: The following tool images are not available locally:[/]")
    for missing_tool_image in missing_tool_images:
        stderr.print("[red]" + missing_tool_image + "[/]")
    typer.confirm("Should DEM try to fix the faulty Development Environment?", abort=True)

    platform.install_dev_env(dev_env_local)
    stdout.print("[green]DEM fixed the " + dev_env_local.name + "![/]")

def execute(platform: Platform, dev_env_name: str, container_arguments: list[str]) -> None:
    """ Execute the run command in the given Dev Env context. If something is wrong with the Dev 
        Env the DEM can try to fix it.

        Args:
            dev_env_name -- name of the Development Environment
            container_arguments -- arguments passed to the container
    """


    dev_env_local = platform.get_dev_env_by_name(dev_env_name)

    if dev_env_local is None:
        stderr.print("[red]Error: Unknown Development Environment: " + dev_env_name + "[/]")
        raise typer.Abort()
    else:
        # Only the local images are needed.
        Platform.local_only = True

        missing_tool_images = set()
        for tool in dev_env_local.tool_image_descriptors:
            tool_image = tool["image_name"] + ":" + tool["image_version"]
            # Check if the required tool image exists locally.
            if tool_image not in platform.tool_images.get_local_ones().keys():
                missing_tool_images.add(tool_image)

        if missing_tool_images:
            handle_missing_tool_images(missing_tool_images, dev_env_local, platform)

        platform.container_engine.run(container_arguments)


"""Direct access to the registry over HTTP. or using the container engine."""
# dem/core/registry.py

from dem.core.core import Core
from dem.core.container_engine import ContainerEngine
from dem.core.exceptions import RegistryError
import requests
from typing import Generator
from abc import ABC, abstractmethod

class Registry(Core, ABC):
    """ Abstract base class for a registry."""
    def __init__(self, container_engine: ContainerEngine, registry_config: dict) -> None:
        """ Init the class.
        
            Args:
                container_engine -- the container engine
                registry_config -- container the name of the registry and its URL
        """
        self._container_engine = container_engine
        self._registry_config = registry_config
        self._repos = []

    @abstractmethod
    def _append_repo_with_tag(self, endpoint_response: dict, repo: str) -> None:
        """ Get the tags from the endpoint response. Save the tags alongside with the actual repo
            in the private repo list.
        """

    @abstractmethod
    def _get_tag_endpoint_url(self, repo_name: str) -> str:
        """ Get the registry specific endpoint to obtain the tags."""

    @abstractmethod
    def _list_repos_in_registry(self) -> Generator:
        """ Generator function for listing the repos. """

    def _list_tags(self, repo: str) -> None:
        """ Get the tags from the respective endpoint and call the registry specific function to 
            populate the private repo list.

            Args:
                repo -- get the tags of this repository
        """
        try:
            response = requests.get(self._get_tag_endpoint_url(repo), 
                                              timeout=self.config_file.http_request_timeout_s)
        except Exception as e:
            self.user_output.error(str(e))
        else:
            if response.status_code == requests.codes.ok:
                self._append_repo_with_tag(response.json(), repo)
                return
            else:
                self.user_output.error("Error in communication with the registry. Failed to retrieve tags. Response status code: " + str(response.status_code))

        self.user_output.msg("Skipping repository: " + repo)
    
    @property
    def repos(self) -> list[str]:
        """ Getter function for the repos in the registry.
        
            Returns with list of the repos.
        """
        self.user_output.status_generator(self._list_repos_in_registry())
        return self._repos

class DockerHub(Registry):
    """ Docker Hub Registry
    
        Class variables:
            _docker_hub_domain -- the Docker Hub domain (used to determine if the config is for a 
                                  Docker Hub registry)
            _tag_endpoint_response_key -- used to obtain the tags from the endpoint response
    """
    _docker_hub_domain = "registry.hub.docker.com"
    _tag_endpoint_response_key = "results"

    def _append_repo_with_tag(self, endpoint_response: dict, repo: str) -> None:
        """ Get the tags from the endpoint response. Save the tags alongside with the actual repo
            in the private repo list.

            Args:
                endpoint_response -- the response from the endpoint
                repo -- append the tags of this repoistory
        """
        for result in endpoint_response[self._tag_endpoint_response_key]:
            self._repos.append(repo + ":" + result["name"])

    def _get_tag_endpoint_url(self, repo: str) -> str:
        """ Get the Docker Hub specific endpoint url to obtain the tags.
            
            Args:
                repo -- we would like to get the tags for this repository
        """
        return self._registry_config["url"] + "/v2/repositories/" + repo + "/tags/"

    def _list_repos_in_registry(self) -> Generator:
        """ Generator function for listing the repos. """
        for repo in self._container_engine.search(self._registry_config["name"]):
            yield "Loading image data from: " + repo
            self._list_tags(repo)

class DockerRegistry(Registry):
    """ Docker Registry
    
        Class variables:
            _tag_endpoint_response_key -- used to obtain the tags from the endpoint response
    """
    _tag_endpoint_response_key = "tags"

    def _append_repo_with_tag(self, endpoint_response: dict, repo: str) -> None:
        """ Get the tags from the endpoint response. Save the tags alongside with the actual repo
            in the private repo list.

            Args:
                endpoint_response -- the response from the endpoint
                repo -- append the tags of this repoistory
        """
        for result in endpoint_response[self._tag_endpoint_response_key]:
            self._repos.append(repo + ":" + result)

    def _get_tag_endpoint_url(self, repo: str) -> str:
        """ Get the Docker Registry specific endpoint url to obtain the tags.
            
            Args:
                repo -- we would like to get the tags for this repository
        """
        return self._registry_config["url"] + "/v2/" + repo.split("/")[1] + "/tags/list"

    def _search(self) -> list[str]:
        """ Search the registry for the repositories
        
            Return with the repo list or an empty list if something bad happened.
        """
        repo_endpoint = self._registry_config["url"] + "/v2/_catalog"

        try:
            response = requests.get(repo_endpoint, timeout=self.config_file.http_request_timeout_s)
        except Exception as e:
            self.user_output.error(str(e))
        else:
            if response.status_code == requests.codes.ok:
                try:
                    return response.json()["repositories"]
                except requests.exceptions.JSONDecodeError as e:
                    self.user_output.error("Invalid JSON format in response. " + str(e))
                except Exception as e:
                    self.user_output.error(str(e))
            else:
                self.user_output.error("Error in communication with the registry. Failed to retrieve the repositories. Response status code: " + str(response.status_code))

        self.user_output.msg("Skipping registry: " + self._registry_config["name"])
        return []

    def _list_repos_in_registry(self) -> Generator:
        """ Generator function for listing the repos. """
        for repo_name in self._search():
            repo = self._registry_config["name"] + '/' + repo_name
            yield "Loading image data from: " + repo
            self._list_tags(repo)

class Registries(Core):
    """ Contains all configured registiries."""
    def __init__(self, container_engine: ContainerEngine) -> None:
        """ Init the class by creating the registry instances.
            
            Args:
                container_engine -- the container engine
                config_file -- config file that contains the registry parameters
        """
        self._container_engine = container_engine
        self.registries: list[Registry] = []
        for registry_config in self.config_file.registries:
            self._add_registry_instance(registry_config)
    
    def _add_registry_instance(self, registry_config: dict) -> None:
        """ Add a registry instance. Currently only the Docker Hub and Docker Registry registry 
            types are supported.
            
            Args:
                registry_config -- registry config
            """
        if DockerHub._docker_hub_domain in registry_config["url"]:
            self.registries.append(DockerHub(self._container_engine, registry_config))
        else:
            self.registries.append(DockerRegistry(self._container_engine, registry_config))

    def list_repos(self, reg_selection: list[str]) -> list[str]:
        """ List the available repositories.

            Args:
                reg_selection -- the selected registries, empty list means all registries
        
            Return with the list of repositories.
        """
        repo_list: list[str] = []

        for registry in self.registries:
            if not reg_selection or registry._registry_config["name"] in reg_selection:
                try:
                    repo_list.extend(registry.repos)
                except Exception as e:
                    self.user_output.error(str(e))
                    self.user_output.error("[red]Error: The " + registry._registry_config["name"] + \
                                           " registry is not available.[/]")

        return repo_list

    def add_registry(self, registry_config: dict) -> None:
        """ Add a new registry.
        
            Args:
                registry -- the registry to add
        """
        self._add_registry_instance(registry_config)

        self.config_file.registries.append(registry_config)
        self.config_file.flush()

    def list_registry_configs(self) -> list[dict]:
        """ List the registry configs. (As stored in the config file.)
        
            Returns with the list of the available registry configurations.
        """
        return self.config_file.registries

    def delete_registry(self, registry_config: dict) -> None:
        """ Delete the registry.
        
            Args:
                registry -- the registry to delete
        """
        for registry in self.registries.copy():
            if registry._registry_config == registry_config:
                self.registries.remove(registry)

        self.config_file.registries.remove(registry_config)
        self.config_file.flush()


"""Development Environment Catalog."""
# dem/core/dev_env_catalog.py

from dem.core.dev_env import DevEnv
from dem.core.core import Core
from dem.core.exceptions import CatalogError
import requests

class DevEnvCatalog(Core):
    """ Development Environment Catalog. """
    def __init__(self, catalog_config: dict) -> None:
        """ Init the class. 

            The name of the catalog must be unique.

            Args:
                catalog_config -- the catalog description
        """
        self.config: dict = catalog_config
        self.url: str = catalog_config["url"]
        self.name: str = catalog_config["name"]
        self.dev_envs: list[DevEnv] = []

    def request_dev_envs(self) -> None:
        """ Request the Development Environments from the catalog. 
        
            Raises:
                CatalogError -- if the communication with the catalog fails
        """
        try:
            deser_json_response: requests.Response = requests.get(self.url, 
                                                                timeout=self.config_file.http_request_timeout_s)
        except Exception as e:
            raise CatalogError(f"Error in communication with the [bold]{self.name}[/bold] Development Environment Catalog.\n{str(e)}")

        if deser_json_response.status_code != requests.codes.ok:
            raise CatalogError(f"Error in communication with the [bold]{self.name}[/bold] Development Environment Catalog. " + 
                               "Failed to retrieve Development Environments." + 
                               "\nResponse status code: " + str(deser_json_response.status_code) + 
                               "\nDoes the URL point to a valid Development Environment Catalog?\n")

        try:
            for dev_env_descriptor in deser_json_response.json()["development_environments"]:
                self.dev_envs.append(DevEnv(descriptor=dev_env_descriptor))
        except Exception as e:
            raise CatalogError(f"The {self.name} Development Environment Catalog is corrupted.\n{str(e)}")

    def get_dev_env_by_name(self, dev_env_name: str) -> DevEnv | None:
        """ Get the Development Environment by name.
        
            Args:
                dev_env_name -- name of the Development Environment to get

            Return with the instance representing the Development Environment. If the Development 
            Environment doesn't exist in the catalog, return with None.
        """
        for dev_env in self.dev_envs:
            if dev_env.name == dev_env_name:
                return dev_env

class DevEnvCatalogs(Core):
    """ List of the available Development Environment Catalogs. """
    def __init__(self) -> None:
        """ Init the class with the catalogs from the config file."""
        self.catalogs: list[DevEnvCatalog] = []
        for catalog_config in self.config_file.catalogs:
            self.catalogs.append(DevEnvCatalog(catalog_config))

    def add_catalog(self, name: str, url:str) -> None:
        """ Add a new catalog.
        
            Args:
                name -- name of the catalog
                url -- url of the catalog

            Raises:
                CatalogError -- if the name is already used
        """
        for catalog in self.catalogs:
            if catalog.name == name:
                raise CatalogError(f"The {name} Development Environment Catalog name is already used.")

        catalog_config = {
            "name": name,
            "url": url
        }
        new_dev_env_catalog = DevEnvCatalog(catalog_config)
        # Request the Development Environments to validate the catalog.
        new_dev_env_catalog.request_dev_envs()

        self.catalogs.append(new_dev_env_catalog)
        self.config_file.catalogs.append(catalog_config)
        self.config_file.flush()

    def delete_catalog(self, name: str) -> None:
        """ Delete the catalog.
        
            Args:
                name -- name of the catalog to delete

            Raises:
                CatalogError -- if the catalog doesn't exist
        """
        for catalog in self.catalogs.copy():
            if catalog.name == name:
                self.catalogs.remove(catalog)
                break
        else:
            raise CatalogError(f"The {name} Development Environment Catalog doesn't exist.")

        self.config_file.catalogs.remove(catalog.config)
        self.config_file.flush()


"""The core can provide information through this module to the user."""
# dem/cli/tui/tui_user_output.py

from dem.cli.console import stdout, stderr
from dem.core.user_output import UserOutput

import typer
from typing import Generator
from rich.progress import Progress, TaskID, TextColumn, BarColumn, TaskProgressColumn
from rich.status import Status

class PullProgressBar():
    """ Visualize the status of the pull command on a progress bar."""
    def __init__(self, generator: Generator) -> None:
        """ Init the class
        
            Args:
                generator -- the status of the pull command is presented through this generator
        """
        self.tasks = {}
        self.generator = generator

    def _update_progress_bar(self, id: str, item: dict, status: str) -> None:
        """ Update the progress bar of the image pull.
        
            Args:
                id -- layer id
                item -- current item from the generator
                status -- which process's status the progress bar shows
        """
        task = self.tasks.get(id)

        if task is None:
            task = self.progress.add_task(str(id), id=id)
            self.tasks[id] = task

        progress_detail = item.get("progressDetail")
        current = None
        total = None
        if progress_detail:
            current = progress_detail.get("current")
            total = progress_detail.get("total")

        if current and total:
            self.progress.update(task, description=str(status), total=float(total), 
                                 completed=float(current))

    def _process(self, item: dict) -> None:
        """ Process an item from the generator provided by the pull command.
        
        Args:
            item-- current item from the generator
            """
        status = item.get("status")
        id = item.get("id")

        if status:
            if id:
                if item.get("progressDetail"):
                    self._update_progress_bar(id, item, status)
                else:
                    self.progress.console.print(str(id) + ": " + str(status))
            else:
                self.progress.console.print(str(status))

    def run_generator(self):
        with Progress(TextColumn("[progress.layer_id]{task.fields[id]}"), 
                      TextColumn("[progress.description]{task.description}"),
                      BarColumn(), TaskProgressColumn(), console=stdout) as self.progress:

            for item in self.generator:
                self._process(item)

class TUIUserOutput(UserOutput):
    """ Provides the interface between the core modules and the rich based TUI."""
    def msg(self, text: str, is_title: bool = False) -> None:
        """ Send a message.
        
            Args:
                text -- the text to print
                is_title -- the text is the title of a new section.
        """
        if is_title is True:
            stdout.rule(text)
        else:
            stdout.print(text)

    def error(self, text: str) -> None:
        """ Send and error message
        
            Args:
                text -- the error message
        """
        stderr.print("[red]" + text + "[/]")

    def get_confirm(self, text: str, confirm_text: str) -> None:
        """ Get confirmation from the user.

            In case the user does not confirm the action, the program gets aborted.
        
            Args:
                text -- message to print (can be empty)
                confirm_text: the action the user needs to confirm
        """
        if text != "":
            stdout.print(text)

        typer.confirm(confirm_text, abort=True)

    def progress_generator(self, generator: Generator) -> None:
        """ Process the progress generator. 

            The input generator must be exhausted.

            Args:
                generator -- the generator
        """
        pull_progress_bar = PullProgressBar(generator)
        pull_progress_bar.run_generator()

    def status_generator(self, generator: Generator) -> None:
        """ Process the status generator. 

            The input generator must be exhausted.

            Args:
                generator -- the generator
        """
        with Status("") as status:
            status.start()

            for item in generator:
                status.update(item)

            status.stop()


"""CLI command implementation for adding a catalog."""
# dem/cli/command/add_cat_cmd.py

from dem.core.platform import Platform
from dem.core.exceptions import CatalogError
from dem.cli.console import stderr, stdout

def execute(platform: Platform, name: str, url:str) -> None:
    """ Add a new Dev Env Catalog.
    
        Args:
            platform -- the platform
            name -- name of the catalog
            url -- URL of the catalog's JSON file
    """
    try:
        platform.dev_env_catalogs.add_catalog(name, url)
        stdout.print("[green]The catalog has been successfully added.[/]")
    except CatalogError as e:
        stderr.print(f"[red]{str(e)}[/]\n")
        stderr.print("[red]The catalog could not be added.[/]")


"""CLI command implementation for deleting a Dev Env Catalog."""
# dem/cli/command/del_cat_cmd.py

from dem.core.platform import Platform
from dem.core.exceptions import CatalogError
from dem.cli.console import stdout, stderr

def execute(platform: Platform, catalog_name: str) -> None:
    """ Delete the Dev Env Catalog.

        Args:
            platform -- the platform
            catalog_name -- name of the catalog to delete
    """
    try:
        platform.dev_env_catalogs.delete_catalog(catalog_name)
        stdout.print(f"[green]The [bold]{catalog_name}[/bold] catalog has been successfully deleted.")
    except CatalogError as e:
        stderr.print(f"[red]{str(e)}[/]\n")
        stderr.print("[red]The catalog could not be deleted.[/]")


"""Docker container engine management."""
# dem/core/container_engine.py

from dem.core.core import Core
from dem.core.exceptions import ContainerEngineError
import docker
import docker.errors

class ContainerEngine(Core):
    """ Operations on the Docker Container Engine."""

    def __init__(self) -> None:
        """ Init the class."""
        self._docker_client = docker.from_env()

    def get_local_tool_images(self) -> list[str]:
        """ Get local tool images.
        
            Return with the list of the locally avialable tool image names.
        """
        local_image_tags = []

        for image in self._docker_client.images.list():
            for tag in image.tags:
                if tag:
                    local_image_tags.append(tag)

        return local_image_tags

    def pull(self, repository: str) -> None:
        """ Pull a repository from the axemsolutions registry.
        
            Args:
                repository -- repository to pull
        """
        resp = self._docker_client.api.pull(repository, stream=True, decode=True)
        self.user_output.progress_generator(resp)

    def run(self, container_arguments: list[str]) -> None:
        """ Run the container. 
        
            The function converts the Docker CLI commands to Docker Engine API call parameters.

            The container always gets started in detach mode. If the -d option is enabled the 
            function returns after the container has been started. If not enabled the DEM streams 
            the logs from the container to the user output while it is running. This effectively 
            results in the same behaviour as the docker run command's -d option.

            Args:
                container_arguments -- list of arguments to pass to the API call
        """
        container_arguments_iter = iter(container_arguments)

        image = ""
        ports = {}
        name = ""
        volumes = []
        command = ""
        privileged = False
        auto_remove = False
        stream_logs = True

        try:
            for argument in container_arguments_iter:
                if argument.startswith("-"):
                    match argument:
                        case "-p":
                            port_binding = next(container_arguments_iter)
                            try:
                                host_port, container_port = port_binding.split(":")
                            except ValueError:
                                raise ContainerEngineError("The option -p has invalid argument: " + port_binding)
                            ports[container_port] = int(host_port)
                        case "--name":
                            name = next(container_arguments_iter)
                        case "-v":
                            volumes.append(next(container_arguments_iter))
                        case "--privileged":
                            privileged = True
                        case "--rm":
                            auto_remove = True
                        case "-d":
                            stream_logs = False
                        case _:
                            raise ContainerEngineError("The input parameter " + argument + " is not supported!")
                else:
                    image = argument
                    for argument in container_arguments_iter:
                        command += argument + " "
                    command = command[:-1]
        except StopIteration:
            raise ContainerEngineError("Invalid input parameter!")

        run_result = self._docker_client.containers.run(image, command=command, 
                                                        auto_remove=auto_remove, 
                                                        privileged=privileged, volumes=volumes,
                                                        ports=ports, name=name, stderr=True, 
                                                        detach=True)

        if stream_logs:
            for line in run_result.logs(stream=True):
                self.user_output.msg(line.decode().strip())

    def remove(self, image: str) -> None:
        """ Remove a tool image.

            If the removal was successful return with True, otherwise return with False.
        
            Args: 
                image -- the tool image to remove
        """
        try:
            self._docker_client.images.remove(image)
        except docker.errors.ImageNotFound:
            self.user_output.msg(f"[yellow]The {image} doesn't exist. Unable to remove it.[/]\n")
        except docker.errors.APIError:
            raise ContainerEngineError(f"The {image} is used by a container. Unable to remove it.\n")
        else:
            self.user_output.msg(f"[green]Successfully removed the {image}![/]\n")

    def search(self, registry: str) -> list[str]:
        """ Search repository in the axemsolutions registry.
        
            Args:
                registry -- registry to search
        """
        local_registryimagelist = []

        for repositories in self._docker_client.images.search(registry):
            local_registryimagelist.append(repositories['name'])

        return local_registryimagelist


"""CLI command implementation for adding a registry."""
# dem/cli/command/add_reg_cmd.py

from dem.core.platform import Platform
from dem.cli.console import stdout

def execute(platform: Platform, name: str, url:str) -> None:
    """ Add a new registry.
    
        Args:
            name -- name or IP address of the registry
            url -- API URL of the registry 
    """
    
    registry = {
        "name": name,
        "url": url
    }
    if registry not in platform.registries.list_registry_configs():
        platform.registries.add_registry(registry)
    else:
        stdout.print("[yellow]The input registry is already added.[/]")


"""Implementation of the init command."""
# dem/cli/command/init_cmd.py

import os, typer
from dem.core.platform import Platform
from dem.core.dev_env import DevEnv
from dem.core.exceptions import PlatformError
from dem.cli.console import stderr, stdout

def execute(platform: Platform, project_path: str) -> None:
    """ Initialize a project at the given path.

        Args:
            platform -- the platform
            project_path -- the path to the project to initialize
    """


    if not os.path.isdir(project_path):
        stderr.print(f"[red]Error: The {project_path} path does not exist.[/]")
        return

    try:
        dev_env = DevEnv(descriptor_path=f"{project_path}/.axem/dev_env_descriptor.json")
    except FileNotFoundError as e:
        stderr.print(f"[red]Error: No Dev Env is assigned to this project. You can assign one with `dem assign`.")
        return

    for local_dev_env in platform.local_dev_envs:
        if local_dev_env.name == dev_env.name:
            stdout.print(f"[yellow]Warning: The {dev_env.name} Development Environment is already initialized.[/]")
            typer.confirm("Would you like to re-init the Dev Env? All local changes will be lost!", abort=True)

            if local_dev_env.is_installed:
                typer.confirm("The Development Environment is installed, so it can't be deleted. Do you want to uninstall it first?", 
                              abort=True)
                
                try:
                    platform.uninstall_dev_env(local_dev_env)
                except PlatformError as e:
                    stderr.print(f"[red]{str(e)}[/]")
                    return
            
            platform.local_dev_envs.remove(local_dev_env)
            break

    platform.local_dev_envs.append(dev_env)
    platform.flush_descriptors()
    stdout.print(f"[green]Successfully initialized the {dev_env.name} Dev Env for the project at {project_path}![/]")
    stdout.print(f"\nNow you can install the Dev Env with the `dem install {dev_env.name}` command.")


"""json file handling."""
# dem/core/data_management.py

from typing import Any
from dem.core.properties import __config_dir_path__
from dem.core.exceptions import DataStorageError
from pathlib import PurePath
import os
import json

class BaseJSON():
    """ This class acts as an abstracted buffer over a json file. 
    
        If the buffer is not up-to-date it can be updated with update() which is a read from the 
        json file. If the buffer has newer data it can be written to file with flush().

        Class attributes: 
            _config_dir -- points to the json files' directory
            _path -- path to the json file (must be set in the descending classes)
            _default_json -- default json content
        """
    _config_dir = os.path.expanduser('~') + __config_dir_path__
    _path = ""
    _default_json = ""

    def _create_default_json(self) -> dict:
        """ If the .json doesn't exist, then create the default one.
        
            Return with the deserialized default json content. 
        """
        is_path_exist = os.path.exists(self._config_dir)
        if not is_path_exist:
            os.makedirs(self._config_dir)

        json_file = open(self._path, "w")
        json_file.write(self._default_json)
        json_file.close()

        return json.loads(self._default_json)

    def __init__(self) -> None:
        """ Init the class with an empty dict for the deserialized dev_env.json file. 
            Later this variable can be used to access the deserialized data. 
        """
        self.deserialized: dict[str, Any] = {}

    def update(self) -> None:
        """ Update the buffer with the content from the json file."""
        try: 
            json_file = open(self._path, "r")
        except FileNotFoundError:
            self.deserialized = self._create_default_json()
        else:
            self.deserialized = json.load(json_file)
            json_file.close()

    def flush(self) -> None:
        """ Write the buffer content to the json file."""
        json_file = open(self._path, "w")
        json.dump(self.deserialized, json_file, indent=4)
        json_file.close()

    def restore(self) -> None:
        """ Restore the json file to its default content."""
        self.deserialized = self._create_default_json()

class LocalDevEnvJSON(BaseJSON):
    """ Serialize and deserialize the dev_env.json file."""
    def __init__(self) -> None:
        """ Init the class."""
        self._path = PurePath(self._config_dir + "/dev_env.json")
        self._default_json = """{
    "version": "0.1",
    "org_name": "axem",
    "registry": "registry-1.docker.io",
    "development_environments": []
}
"""
        super().__init__()

    def update(self) -> None:
        try:
            super().update()
        except json.decoder.JSONDecodeError as e:
            raise DataStorageError(f"The dev_env.json file is corrupted.\n{str(e)}") from e

class ConfigFile(BaseJSON):
    """ Serialize and deserialize the config.json file."""
    def __init__(self) -> None:
        """ Init the class."""
        self._path = PurePath(self._config_dir + "/config.json")
        self._default_json = """{
    "registries": [
        {
            "name": "axemsolutions",
            "url": "https://registry.hub.docker.com"
        }
    ],
    "catalogs": [
        {
            "name": "axem",
            "url": "https://axemsolutions.io/dem/dev_env_org.json"
        }
    ],
    "hosts": [],
    "http_request_timeout_s": 2
}"""
        super().__init__()

    def update(self) -> None:
        try:
            super().update()
        except json.decoder.JSONDecodeError as e:
            raise DataStorageError(f"The config.json file is corrupted.\n{str(e)}") from e

        self.registries: list[dict] = self.deserialized.get("registries", [])
        self.catalogs: list[dict] = self.deserialized.get("catalogs", [])
        self.hosts: list[dict] = self.deserialized.get("hosts", [])
        self.http_request_timeout_s: float = self.deserialized.get("http_request_timeout_s", None)
        
        if self.http_request_timeout_s is None:
            raise DataStorageError("The http_request_timeout_s is not set in the config.json file.")


"""This module provides the CLI."""
# dem/cli/main.py

import typer, importlib.metadata
from typing import Generator
from typing_extensions import Annotated
import os
from dem import __command__, __app_name__
from dem.cli.command import cp_cmd, import_cmd, info_cmd, list_cmd, create_cmd, modify_cmd, delete_cmd, \
                            rename_cmd, run_cmd, export_cmd, clone_cmd, add_reg_cmd, \
                            list_reg_cmd, del_reg_cmd, add_cat_cmd, list_cat_cmd, del_cat_cmd, \
                            add_host_cmd, uninstall_cmd, install_cmd, assign_cmd, init_cmd, \
                            list_host_cmd, del_host_cmd, list_tools_cmd
from dem.cli.console import stdout
from dem.core.platform import Platform
from dem.core.exceptions import InternalError

typer_cli: typer.Typer = typer.Typer(rich_markup_mode="rich")
platform: Platform | None = None

# Autocomplete functions
def autocomplete_dev_env_name(incomplete: str) -> Generator:
    """ 
    Autocomplete the input Dev Env name with the available matching local Dev Envs.

    Return with the matching Dev Env names by a Generator.
    
    Args:
        incomplete -- the parameter the user supplied so far when the tab was pressed
    """
    if platform is not None:
        for dev_env in platform.local_dev_envs:
            if dev_env.name.startswith(incomplete) or (incomplete == ""):
                yield dev_env.name

def autocomplete_cat_name(incomplete: str) -> Generator:
    """ 
    Autocomplete the input Catalog name with the available matching catalogs.

    Return with the matching Catalog name by a Generator.
    
    Args:
        incomplete -- the parameter the user supplied so far when the tab was pressed
    """
    if platform is not None:
        for catalog in platform.dev_env_catalogs.catalogs:
            if catalog.name.startswith(incomplete) or (incomplete == ""):
                yield catalog.name

def autocomplete_reg_name(incomplete: str) -> Generator:
    """ 
    Autocomplete the input Registry name with the available matching Registries.

    Return with the matching Registry name by a Generator.
    
    Args:
        incomplete -- the parameter the user supplied so far when the tab was pressed
    """
    if platform is not None:
        for registry_config in platform.registries.list_registry_configs():
            if registry_config["name"].startswith(incomplete) or (incomplete == ""):
                yield registry_config["name"]

def autocomplete_host_name(incomplete: str) -> Generator:
    """ 
    Autocomplete the input Host name with the available matching Hosts.

    Return with the matching Host name by a Generator.
    
    Args:
        incomplete -- the parameter the user supplied so far when the tab was pressed
    """
    if platform is not None:
        for host_config in platform.hosts.list_host_configs():
            if host_config["name"].startswith(incomplete) or (incomplete == ""):
                yield host_config["name"]

# DEM commands
@typer_cli.command("list", context_settings={"allow_extra_args": True}) # "list" is a Python keyword
def list_(cat: Annotated[bool, typer.Option(help="List the Dev Envs available from the catalogs.",
                                            show_default=False)] = False,
          ctx: Annotated[typer.Context, typer.Option()] = None) -> None:
    """
    List the Dev Envs.

    By default the local Dev Envs will be listed.
    
    --cat: List the available Dev Envs from the catalogs. Specify the catalogs' name to list the Dev 
    Envs from. More then one catalog can be specified. If no catalog is specified, all the available
    catalogs will be used.
    """
    if platform and ctx:
        list_cmd.execute(platform, cat, ctx.args)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command(context_settings={"allow_extra_args": True})
def list_tools(reg: Annotated[bool, typer.Option(help="List the available tools in the registries.",
                                                  show_default=False)] = False,
               ctx: Annotated[typer.Context, typer.Option()] = None) -> None:
    """
    List the available tools.

    By default the local tools will be listed.

    --reg: List the available tools in the registries. Specify the registry's name to list the tools
    from. More then one registry can be specified. If no registry is specified, all the available
    registries will be used.
    """
    if platform and ctx:
        list_tools_cmd.execute(platform, reg, ctx.args)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command(context_settings={"allow_extra_args": True})
def info(dev_env_name: Annotated[str, typer.Argument(help="Name of the Development Environment to get info about.",
                                                     autocompletion=autocomplete_dev_env_name)],
         cat: Annotated[bool, typer.Option(help="Get the Dev Env from the catalogs")] = False,
         ctx: Annotated[typer.Context, typer.Option()] = None) -> None:
    """
    Get information about the specified Development Environment available locally or in the catalogs.

    --cat: DEM will search for the Dev Env in the catalogs and will print the details of the first
    match. You can specifiy the catalogs' name to search in after this option. If no catalog is 
    specified, all the available catalogs will be used. If the Dev Env is not found in the catalogs,
    an error message will be printed.

    Note: Autocomplete only works with the locally avialable Dev Envs.
    """
    if platform and ctx:
        info_cmd.execute(platform, dev_env_name, cat, ctx.args)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command()
def cp(dev_env_name: Annotated[str, typer.Argument(help="Name of the Development Environment to cp.",
                                                   autocompletion=autocomplete_dev_env_name)],
       new_dev_env_name: Annotated[str, typer.Argument(help="Name of the New Development Environment.")]) -> None:
    """
    Create a copy of a local Dev Env.
    """
    if platform:
        cp_cmd.execute(platform, dev_env_name, new_dev_env_name)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command()
def create(dev_env_name: Annotated[str, typer.Argument(help="Name of the Development Environment to create.")]) -> None:
    """
    Create a new Development Environment.
    """
    if platform:
        create_cmd.execute(platform, dev_env_name)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command()
def export(dev_env_name: Annotated[str, typer.Argument(help="Name of the Development Environment to export.",
                                                       autocompletion=autocomplete_dev_env_name)],
           path_to_export: Annotated[str, typer.Argument(help="Path where to extract the Dev Env.")] = "") -> None:
    """
    Export the Development Environment.
    """
    if platform:
        export_cmd.execute(platform, dev_env_name,path_to_export)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command("import")
def import_(path_to_dev_env: Annotated[str, typer.Argument(help="Path to the Dev Env to import.")]) -> None:
    """
    Import the Development Environment.
    """
    if platform:
        import_cmd.execute(platform, path_to_dev_env)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command()
def clone(dev_env_name: Annotated[str, typer.Argument(help="Name of the Dev Env descriptor to clone.")]) -> None:
    """
    Copy the Dev Env's descriptor from the catalog to the local descriptor storage.
    """
    if platform:
        clone_cmd.execute(platform, dev_env_name)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command()
def rename(dev_env_name: Annotated[str, typer.Argument(help="Name of the Development Environment to rename.",
                                                       autocompletion=autocomplete_dev_env_name)],
           new_dev_env_name: Annotated[str, typer.Argument(help="The new name.")]) -> None:
    """
    Rename the Development Environment.
    """
    if platform:
        rename_cmd.execute(platform, dev_env_name,new_dev_env_name)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command()
def modify(dev_env_name: Annotated[str, typer.Argument(help="Name of the Development Environment to modify.",
                                                       autocompletion=autocomplete_dev_env_name)]) -> None:
    """
    Change a tool in a Development Environment.

    If the tool type is not specified, the Dev Env settings panel will be opened.
    """
    if platform:
        modify_cmd.execute(platform, dev_env_name)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command()
def delete(dev_env_name: Annotated[str, typer.Argument(help="Name of the Development Environment to delete.",
                                                       autocompletion=autocomplete_dev_env_name)]) -> None:
    """
    Delete the Dev Env descriptor from the local descriptor storage.
    If the Dev Env is installed, the user will be asked whether they want to uninstall it. 
    """
    if platform:
        delete_cmd.execute(platform, dev_env_name)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command()
def install(dev_env_name: Annotated[str, typer.Argument(help="Name of the Development Environment to install.",
                                                       autocompletion=autocomplete_dev_env_name)]) -> None:
    """
    Install the Development Environment.
    """
    if platform is not None:
        install_cmd.execute(platform, dev_env_name)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")
    
@typer_cli.command()
def uninstall(dev_env_name: Annotated[str, typer.Argument(help="Name of the Development Environment to uninstall.",
                                                       autocompletion=autocomplete_dev_env_name)]) -> None:
    """
    Uninstall the Development Environment from the local setup. If a tool image is not required
    anymore by any of the available local Development Environments, the DEM will delete it.
    """
    if platform:
        uninstall_cmd.execute(platform, dev_env_name)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")  

@typer_cli.command()
def assign(dev_env_name: Annotated[str, typer.Argument(help="Name of the Dev Env that should be assign to the project.",
                                                       autocompletion=autocomplete_dev_env_name)],
           project_path: Annotated[str, typer.Argument(help="Path of the project.")] = os.getcwd()) -> None:
    """
    Assign a Development Environment to a project.

    If the project path is not specified, the current working directory will be used.
    """
    if platform:
        assign_cmd.execute(platform, dev_env_name, project_path)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command()
def init(project_path: Annotated[str, typer.Argument(help="Path of the project.")] = os.getcwd()) -> None:
    """
    Initialize a project to use a Development Environment.

    If the project path is not specified, the current working directory will be used.
    """
    if platform:
        init_cmd.execute(platform, project_path)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
def run(dev_env_name: Annotated[str, typer.Argument(help="Run the container in this Development Environment context",
                                                    autocompletion=autocomplete_dev_env_name)],
        ctx: Annotated[typer.Context, typer.Option()]) -> None:
    """
    Run the `docker run` command in the Development Environment's context with the given parameters.  

    This command can be used as the docker CLI one, except as first argument the name of the 
    Development Environment must be set. 
    Example: dem run dev_env --name test test_image_name:latest ls -la

    See the documentation for the list of currently supported docker run parameters.
    """
    if platform:
        run_cmd.execute(platform, dev_env_name, ctx.args)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command()
def add_reg(name: Annotated[str, typer.Argument(help="Name of the registry to add")], 
            url: Annotated[str, typer.Argument(help="API URL of the registry")]) -> None:
    """
    Add a new registry.

    The name of the registry is what you would normally use to pull an image. 
    Examples:
        - If the full image tag: repository/image:tag -> the name should be repository.
        - If the full image tag: 192.168.1.1:5000/image:tag -> the name should be 192.168.1.1:5000

    The URL should point to the registry's REST API. For the Docker Hub its 
    https://registry.hub.docker.com, or it can be http://localhost:5000 for a self-hosted one.
    """
    if platform:
        add_reg_cmd.execute(platform, name, url)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command()
def list_reg() -> None:
    """
    List the available registries.
    """
    if platform:
        list_reg_cmd.execute(platform)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command()
def del_reg(registry_name: Annotated[str, typer.Argument(help="Name or IP address of the registry to delete.",
                                                         autocompletion=autocomplete_reg_name)]) -> None:
    """
    Delete a registry.
    """
    if platform:
        del_reg_cmd.execute(platform, registry_name)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command()
def add_cat(name: Annotated[str, typer.Argument(help="Name of the Development Environment Catalog to add")], 
            url: Annotated[str, typer.Argument(help="URL of the Development Environment Catalog's JSON file")]) -> None:
    """
    Add a new catalog.

    You can name the catalog as you wish. 
    
    The URL must point to an HTTP(S) server where the Catalog json file is available.
    """
    if platform:
        add_cat_cmd.execute(platform, name, url)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command()
def list_cat() -> None:
    """
    List the available catalogs.
    """
    if platform:
        list_cat_cmd.execute(platform)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command()
def del_cat(catalog_name: Annotated[str, typer.Argument(help="Name of the Development Environment Catalog to delete.",
                                                        autocompletion=autocomplete_cat_name)]) -> None:
    """
    Delete a catalog.
    """
    if platform:
        del_cat_cmd.execute(platform, catalog_name)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command()
def add_host(name: Annotated[str, typer.Argument(help="Name of the host")],
             address: Annotated[str, typer.Argument(help="IP or hostname of the host")]) -> None:
    """
    Add a new host.
    """
    if platform:
        add_host_cmd.execute(platform, name, address)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command()
def list_host() -> None:
    """
    List the available hosts.
    """
    if platform is not None:
        list_host_cmd.execute(platform)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

@typer_cli.command()
def del_host(host_name: Annotated[str, typer.Argument(help="Name of the host to delete.", 
                                                      autocompletion=autocomplete_host_name)]) -> None:
    """
    Delete a host.
    """
    if platform:
        del_host_cmd.execute(platform, host_name)
    else:
        raise InternalError("Error: The platform hasn't been initialized properly!")

def _version_callback(value: bool) -> None:
    if value:
        try: 
            version = importlib.metadata.version(__app_name__)
        except importlib.metadata.PackageNotFoundError:
            stdout.print("[yellow]Install DEM to get the version number.[/]")
        else: 
            stdout.print("[cyan]" + __app_name__ + " v" + version + "[/]")

        raise typer.Exit()

@typer_cli.callback()
def main(
    version: bool = typer.Option(
        None,
        "--version",
        "-v",
        help="Show the dem version.",
        callback=_version_callback,
        is_eager=True,
    )) -> None:
    """
    Development Environment Manager (dem)
    
    Manage your containerized Development Environments with ease.
    """
    return



"""CLI command implementation for listing the available Development Environment Catalogs."""
# dem/cli/command/list_cat_cmd.py

from dem.core.platform import Platform
from dem.core.exceptions import CatalogError
from dem.cli.console import stdout, stderr
from rich.table import Table

def execute(platform: Platform) -> None:
    """ List available Development Environment Catalogs.
    
        Args:
            platform -- the platform
    """
    table = Table()
    table.add_column("name")
    table.add_column("url")

    if not platform.dev_env_catalogs.catalogs:
        stdout.print("[yellow]No Development Environment Catalogs are available![/]")
        return

    for catalog in platform.dev_env_catalogs.catalogs:
        try:
            catalog.request_dev_envs()
        except CatalogError as e:
            stderr.print(f"[red]{str(e)}[/]")

        table.add_row(catalog.name, catalog.url)

    stdout.print(table)


"""create CLI command implementation."""
# dem/cli/command/create_cmd.py

import typer
from dem.core.dev_env import DevEnv
from dem.core.tool_images import ToolImage
from dem.core.platform import Platform
from dem.core.exceptions import PlatformError
from dem.cli.console import stdout, stderr
from dem.cli.tui.window.dev_env_settings_window import DevEnvSettingsWindow
from dem.cli.tui.printable_tool_image import convert_to_printable_tool_images

def open_dev_env_settings_panel(all_tool_images: dict[str, ToolImage]) -> list[str]:
    """ Open the Development Environment settings panel.
        
        Args:
            all_tool_images -- the Tool Images

        Returns:
            the selected Tool Image names
    """
    dev_env_settings_panel = DevEnvSettingsWindow(convert_to_printable_tool_images(all_tool_images))
    dev_env_settings_panel.wait_for_user()

    if "cancel" in dev_env_settings_panel.cancel_save_menu.get_selection():
        raise typer.Abort()

    return dev_env_settings_panel.tool_image_menu.get_selected_tool_images()

def create_new_dev_env_descriptor(dev_env_name: str, selected_tool_images: list[str]) -> dict:
    """ Create a new Development Environment descriptor.
        
        Args:
            dev_env_name -- the name of the Development Environment
            selected_tool_images -- the selected Tool Images

        Returns:
            the descriptor of the new Development Environment
    """
    dev_env_descriptor = {
        "name": dev_env_name,
        "tools": []
    }

    for tool_image in selected_tool_images:
        if "/" in tool_image:
            registry, image = tool_image.split("/")
            image_name = registry + '/' + image.split(":")[0]
        else:
            image = tool_image
            image_name = image.split(":")[0]

        tool_descriptor = {
            "image_name": image_name,
            "image_version": image.split(":")[1]
        }
        dev_env_descriptor["tools"].append(tool_descriptor)

    return dev_env_descriptor

def create_new_dev_env(platform: Platform, new_dev_env_descriptor: dict) -> None:
    """ Create a new Development Environment.
        
        Args:
            platform -- the platform
            new_dev_env_descriptor -- the descriptor of the new Development Environment
    """
    dev_env = DevEnv(descriptor=new_dev_env_descriptor)
    dev_env.assign_tool_image_instances(platform.tool_images)
    platform.local_dev_envs.append(dev_env)

def create_dev_env(platform: Platform, dev_env_name: str) -> None:
    """ Create a new Development Environment or overwrite an existing one.
        
        Args:
            platform -- the platform
            dev_env_name -- the name of the Development Environment

        Exceptions:
            Abort -- if the name of the Development Environment contains whitespace characters
    """
    if ' ' in dev_env_name:
        stderr.print("The name of the Development Environment cannot contain whitespace characters!")
        raise typer.Abort()

    dev_env_original = platform.get_dev_env_by_name(dev_env_name)
    if dev_env_original is not None:
        typer.confirm("The input name is already used by a Development Environment. Overwrite it?", 
                      abort=True)

        if dev_env_original.is_installed:
            typer.confirm("The Development Environment is installed, so it can't be overwritten. " + \
                          "Uninstall it first?", abort=True)
            try:
                platform.uninstall_dev_env(dev_env_original)
            except PlatformError as e:
                stderr.print(f"[red]{str(e)}[/]")
                raise typer.Abort()

    selected_tool_images = open_dev_env_settings_panel(platform.tool_images.all_tool_images)
    new_dev_env_descriptor = create_new_dev_env_descriptor(dev_env_name, selected_tool_images)
    
    if dev_env_original is not None:
        dev_env_original.tool_image_descriptors = new_dev_env_descriptor["tools"]
    else:
        create_new_dev_env(platform, new_dev_env_descriptor)

def execute(platform: Platform, dev_env_name: str) -> None:
    """ Create a new Development Environment.
        
        Args:
            platform -- the platform
            dev_env_name -- the name of the Development Environment

        Exceptions:
            Abort -- if the name of the Development Environment contains whitespace characters
    """
    platform.assign_tool_image_instances_to_all_dev_envs()

    create_dev_env(platform, dev_env_name)
    platform.flush_descriptors()
    stdout.print(f"The [green]{dev_env_name}[/] Development Environment has been created!")
    stdout.print("Run [italic]dem install[/] to install it.")


"""cp CLI command implementation."""
# dem/cli/command/rename_cmd.py

import copy
from dem.core.dev_env import DevEnv
from dem.core.platform import Platform
from dem.cli.console import stderr

def get_dev_env_to_cp(platform: Platform, dev_env_name: str) -> (DevEnv | None):
    dev_env_to_cp = platform.get_dev_env_by_name(dev_env_name) 

    if dev_env_to_cp is None:
        stderr.print("[red]Error: The input Development Environment does not exist.[/]")
        return None
    else:
        return dev_env_to_cp

def check_new_dev_env_name_taken(dev_env_local_setup: Platform, new_dev_env_name: str) -> bool:
    if dev_env_local_setup.get_dev_env_by_name(new_dev_env_name):
        stderr.print("[red]Error: Development environment already exists with the " + new_dev_env_name + " name.[/]")
        return True
    else:
        return False

def cp_given_dev_env(platform: Platform, dev_env_to_cp: DevEnv,
                        new_dev_env_name: str):
    new_dev_env = copy.deepcopy(dev_env_to_cp)
    new_dev_env.name = new_dev_env_name
    new_dev_env.is_installed = False
    platform.local_dev_envs.append(new_dev_env)
    platform.flush_descriptors()

def execute(platform: Platform, dev_env_to_cp_name: str, new_dev_env_name: str) -> None:


    dev_env_to_cp = get_dev_env_to_cp(platform, dev_env_to_cp_name)

    if (dev_env_to_cp is not None and
        check_new_dev_env_name_taken(platform,new_dev_env_name) is False):
        cp_given_dev_env(platform, dev_env_to_cp, new_dev_env_name)


"""CLI command implementation for deleting a Host."""
# dem/cli/command/del_host_cmd.py

from dem.core.platform import Platform
from dem.cli.console import stdout, stderr

def execute(platform: Platform, host_name: str) -> None:
    """ Delete the Host.

        Args:
            host_name -- name of the host to delete
    """
    for host_config in platform.hosts.list_host_configs():
        if host_config["name"] == host_name:
            platform.hosts.delete_host(host_config)
            stdout.print("[green]Host deleted successfully![/]")
            break
    else:
        stderr.print("[red]Error: The input Host does not exist.[/]")


"""The core can provide information through this module to the user."""
# dem/core/user_output.py

from abc import ABC, abstractmethod
from typing import Generator

class UserOutput(ABC):
    """ Abstract base class for the user output. Acts as an interface between the core modules and 
        the UI.
    """
    @abstractmethod
    def msg(self, text: str, is_title: bool = False) -> None:
        """ Send a message.
        
            Args:
                text -- the text to print
                is_title -- the text is the title of a new section.
        """
        pass

    @abstractmethod
    def error(self, text: str) -> None:
        """ Send and error message
        
            Args:
                text -- the error message
        """
        pass

    @abstractmethod
    def get_confirm(self, text: str, confirm_text: str) -> None:
        """ Get confirmation from the user.
        
            Args:
                text -- message to print (can be empty)
                confirm_text: the action the user needs to confirm
        """
        pass

    @abstractmethod
    def progress_generator(self, generator: Generator) -> None:
        """ Process the progress generator. 

            The input generator must be exhausted.

            Args:
                generator -- the generator
        """
        pass

    @abstractmethod
    def status_generator(self, generator: Generator) -> None:
        """ Process the status generator. 

            The input generator must be exhausted.

            Args:
                generator -- the generator
        """
        pass

class NoUserOutput(UserOutput):
    """ This class is assigned to the interface when no UI is present. The methods don't do anything
        except exhausting the generator if applicable.
    """
    def msg(self, text: str, is_title: bool = False) -> None:
        pass

    def error(self, text: str) -> None:
        pass

    def get_confirm(self, text: str, confirm_text: str) -> None:
        pass

    def progress_generator(self, generator: Generator) -> None:
        for _ in generator:
            pass

    def status_generator(self, generator: Generator) -> None:
        for _ in generator:
            pass


"""list CLI command implementation."""
# dem/cli/list_cmd.py

from dem.core.platform import Platform
from dem.core.dev_env import DevEnv
from dem.core.dev_env_catalog import DevEnvCatalog
from dem.cli.console import stdout, stderr
from rich.table import Table
from typing import List

def add_dev_env_info_to_table(platform: Platform, table: Table, dev_env: DevEnv) -> None:
    """ Add Development Environment information to the table.
    
        Args:
            platform -- the Platform
            table -- the Table
            dev_env -- the Development Environment
    """
    if dev_env.is_installed:
        dev_env.assign_tool_image_instances(platform.tool_images)
        installed_column = "[green]Yes[/]"
        tool_image_status = dev_env.get_tool_image_status()
        if tool_image_status == DevEnv.Status.UNAVAILABLE_IMAGE:
            status_column = "[red]Error: Required image is not available![/]"
        elif tool_image_status == DevEnv.Status.REINSTALL_NEEDED:
            status_column = "[red]Error: Incomplete local install![/]"
        else:
            status_column = "Ok"
    else:
        installed_column = "No"
        status_column = "Ok"
    table.add_row(dev_env.name, installed_column, status_column)

def list_local_dev_envs(platform: Platform) -> None:
    """ List the local Development Environments.
    
        Args:
            platform -- the Platform
    """
    if not platform.local_dev_envs:
        stdout.print("[yellow]No Development Environment descriptors are available.[/]")
        return
    else:
        table = Table()
        table.add_column("Name")
        table.add_column("Installed")
        table.add_column("Status")

        for dev_env in sorted(platform.local_dev_envs, key=lambda dev_env: dev_env.name.lower()):
            add_dev_env_info_to_table(platform, table, dev_env)

        stdout.print(f"\n [italic]Local Development Environments[/]")
        stdout.print(table)

def list_actual_cat_dev_envs(catalog: DevEnvCatalog) -> None:
    """ List the Development Environments in the catalog.
    
        Args:
            catalog -- the Development Environment Catalog
    """
    catalog.request_dev_envs()
    if not catalog.dev_envs:
        stdout.print(f"[yellow]No Development Environments are available in the {catalog.name} catalog.[/]")
    else:
        table = Table()
        table.add_column("Name")
        for dev_env in catalog.dev_envs:
            table.add_row(dev_env.name)
        stdout.print(f"\n [italic]Development Environments in the {catalog.name} catalog:[/]")
        stdout.print(table)
    
def list_all_cat_dev_envs(platform: Platform) -> None:
    """ List all Development Environments in the catalogs.
    
        Args:
            platform -- the Platform
    """
    if not platform.dev_env_catalogs.catalogs:
        stdout.print("[yellow]No Development Environment Catalogs are available!")
        return
        
    for catalog in platform.dev_env_catalogs.catalogs:
        list_actual_cat_dev_envs(catalog)

def list_selected_cat_dev_envs(platform: Platform, selected_cats: List[str]) -> None:
    """ List the Development Environments from the specified catalogs.
    
        Args:
            platform -- the Platform
            selected_cats -- the specified catalogs
    """
    for cat_name in selected_cats:
        for catalog in platform.dev_env_catalogs.catalogs:
            if catalog.name == cat_name:
                list_actual_cat_dev_envs(catalog)
                break
        else:
            stderr.print(f"[red]Error: Catalog '{cat_name}' not found![/]")

def execute(platform: Platform, cat: bool, selected_cats: list[str]) -> None:
    """ List Development Environments
    
    Args:
        cat -- if true list all Development Environments in the catalogs
        selected_cats -- list the Development Environments from the specified catalogs
    """
    if cat and not selected_cats:
        list_all_cat_dev_envs(platform)
    elif selected_cats:
        list_selected_cat_dev_envs(platform, selected_cats)
    else:
        list_local_dev_envs(platform)


"""delete CLI command implementation."""
# dem/cli/command/delete_cmd.py

from dem.core.platform import Platform
from dem.core.dev_env import DevEnv
from dem.core.exceptions import PlatformError
from dem.cli.console import stderr, stdout
import typer

def execute(platform: Platform, dev_env_name: str) -> None:


    dev_env_to_delete: DevEnv | None = platform.get_dev_env_by_name(dev_env_name)

    if dev_env_to_delete is None:
        stderr.print(f"[red]Error: The [bold]{dev_env_name}[/bold] Development Environment doesn't exist.")
    else:
        if dev_env_to_delete.is_installed:
            typer.confirm("The Development Environment is installed. Do you want to uninstall it?", 
                          abort=True)

            try:
                platform.uninstall_dev_env(dev_env_to_delete)
            except PlatformError as e:
                stderr.print(f"[red]{str(e)}[/]")
                return

        stdout.print("Deleting the Development Environment...")
        platform.local_dev_envs.remove(dev_env_to_delete)
        platform.flush_descriptors()
        stdout.print(f"[green]Successfully deleted the {dev_env_name}![/]")


"""Local and registry tool images."""
# dem/core/tool_images.py

from dem.core.container_engine import ContainerEngine
from dem.core.registry import Registries
from dem.core.exceptions import ToolImageError

class ToolImage():
    (
        LOCAL_ONLY,
        REGISTRY_ONLY,
        LOCAL_AND_REGISTRY,
        NOT_AVAILABLE,
    ) = range(4)

    def __init__(self, name: str) -> None:
        """ Init the class.
        
            Args:
                name -- the name of the tool image
                
            Exceptions:
                ToolImageError -- if the name is invalid
        """
        self.name = name
        try:
            self.repository = self.name.split(":")[0]
            self.tag = self.name.split(":")[1]
        except IndexError:
            raise ToolImageError(f"Invalid tool image name: {name}")
        self.availability = self.NOT_AVAILABLE

class ToolImages():
    """ Available tool images."""
    def __init__(self, container_engine: ContainerEngine, registries: Registries) -> None:
        """ Init the class.

            The tool images will be obtained by running the update method.
        
            Args: 
                container_engine -- container engine
                registries -- registries
                update_on_instantiation -- set to false for manual update
        """
        self.container_engine = container_engine
        self.registries = registries
        self.all_tool_images = {}

    def update(self, local_only: bool = False, registry_only: bool = False, 
               reg_selection: list[str] = []) -> None:
        """ Update the list of available tools.
        
            Args:
                local_only -- update the local tools only
                registry_only -- update the registry tools only
        """
        registry_tool_image_names = []
        local_tool_image_names = []

        if not registry_only:
            local_tool_image_names = self.container_engine.get_local_tool_images()

        if not local_only:
            registry_tool_image_names = self.registries.list_repos(reg_selection)

        for tool_image_name in local_tool_image_names:
            tool_image = ToolImage(tool_image_name)
            if tool_image_name in registry_tool_image_names:
                tool_image.availability = ToolImage.LOCAL_AND_REGISTRY
            else:
                tool_image.availability = ToolImage.LOCAL_ONLY
            self.all_tool_images[tool_image_name] = tool_image

        for tool_image_name in registry_tool_image_names:
            if tool_image_name not in local_tool_image_names:
                tool_image = ToolImage(tool_image_name)
                tool_image.availability = ToolImage.REGISTRY_ONLY
                self.all_tool_images[tool_image_name] = tool_image

    def get_local_ones(self) -> dict[str, ToolImage]:
        """ Get the local tool images.
        
            Return with the local tool images.
        """
        local_tool_images = {}
        for key, tool_image in self.all_tool_images.items():
            if ((tool_image.availability == ToolImage.LOCAL_ONLY) or 
                (tool_image.availability == ToolImage.LOCAL_AND_REGISTRY)):
                local_tool_images[key] = tool_image

        return local_tool_images
    
    def get_registry_ones(self) -> dict[str, ToolImage]:
        """ Get the registry tool images.
        
            Return with the registry tool images.
        """
        registry_tool_images = {}
        for key, tool_image in self.all_tool_images.items():
            if ((tool_image.availability == ToolImage.REGISTRY_ONLY) or 
                (tool_image.availability == ToolImage.LOCAL_AND_REGISTRY)):
                registry_tool_images[key] = tool_image

        return registry_tool_images


"""install CLI command implementation."""
# dem/cli/command/install_cmd.py

from dem.core.dev_env import DevEnv
from dem.core.platform import Platform, PlatformError
from dem.cli.console import stderr, stdout

def execute(platform: Platform, dev_env_name: str) -> None:
    """
        Install the given Development Environment.
        
        Args:
            platform -- the platform
            dev_env_name -- the name of the Development Environment to install
    """

    platform.assign_tool_image_instances_to_all_dev_envs()

    dev_env_to_install: DevEnv | None = platform.get_dev_env_by_name(dev_env_name)

    if dev_env_to_install is None:
        stderr.print(f"[red]Error: The {dev_env_name} Development Environment does not exist.[/]")
    elif dev_env_to_install.is_installed == True:
        stderr.print(f"[red]Error: The {dev_env_name} Development Environment is already installed.[/]")
    else:
        try:
            platform.install_dev_env(dev_env_to_install)            
        except PlatformError as e:
            stderr.print(f"[red]{e}[/]")
        else:
            stdout.print(f"[green]Successfully installed the {dev_env_name}![/]")


"""CLI command implementation for listing the available hosts."""
# dem/cli/command/list_host_cmd.py

from dem.core.platform import Platform
from dem.cli.console import stdout
from rich.table import Table

def execute(platform: Platform) -> None:
    """ List available Hosts.
    
    Usage: dem list-host
    
    """

    hosts: list[dict] = platform.hosts.list_host_configs()
    """
    Ideally, if 'hosts' is populated, it should look something like:
    hosts = [
                {'name': 'host1', 'address': 'ip_address1'}
                {'name': 'host2', 'address': 'ip_address2'}
                .... and so on
            ]
    if there are no hosts:
    hosts = [  ]
    """
    table = Table()
    table.add_column("name")
    table.add_column("address")

    if not hosts:
        stdout.print("[yellow]No available remote hosts![/]")
    else:
        for host in hosts:
            table.add_row(host['name'], host['address'])

        stdout.print(table)


"""import CLI command implementation."""
# dem/cli/command/import_cmd.py

from dem.core.dev_env import DevEnv
from dem.core.platform import Platform
from dem.cli.console import stderr
import json, os
import typer

def import_dev_env_from_json(platform: Platform,path_to_dev_env: str) -> None:
    """ Import a Development Environment from a JSON file.
    
        Args:
            platform (Platform): The platform instance.
            path_to_dev_env (str): The path to the JSON file.
            
        Raises:
            typer.Abort: If the Development Environment already exists, or the JSON format is 
            invalid.
    """
    raw_file = open(path_to_dev_env, "r")   
            
    try:
        dev_env_descriptor_to_import = json.load(raw_file)
        
        if platform.get_dev_env_by_name(dev_env_descriptor_to_import["name"]) is not None:
            stderr.print("[red]Error: The Development Environment already exists.[/]")
            raise typer.Abort()
        else:        
            new_dev_env: DevEnv = DevEnv(dev_env_descriptor_to_import)
            platform.local_dev_envs.append(new_dev_env)
    except json.decoder.JSONDecodeError:
       stderr.print("[red]Error: invalid json format.[/]")
       raise typer.Abort()
    else:
        raw_file.close()

def execute(platform: Platform, path_to_dev_env: str) -> None:
    """ Execute the import command.
    
        Args:
            platform (Platform): The platform instance.
            path_to_dev_env (str): The path to the JSON file.
            
        Raises:
            typer.Abort: If the Development Environment already exists, or the JSON format is 
            invalid.
    """
    if os.path.exists(path_to_dev_env):
        import_dev_env_from_json(platform,path_to_dev_env)        
        platform.flush_descriptors()
    else:
        stderr.print("[red]Error: The input file does not exist.[/]")


"""Entry point for dem."""
# dem/__main__.py

from dem import __command__
from dem.cli.console import stderr, stdout
from dem.core.exceptions import RegistryError, ContainerEngineError, InternalError, DataStorageError, \
                                CatalogError, ToolImageError
import dem.cli.main
from dem.core.core import Core
from dem.core.platform import Platform
from dem.cli.tui.tui_user_output import TUIUserOutput
import docker.errors
import typer

def main() -> None:
    """ Entry point for the CLI application"""

    # Create the Development Platform
    dem.cli.main.platform = Platform()

    # Connect the UI to the user output interface
    Core.set_user_output(TUIUserOutput())

    try:
        # Load the configuration file
        dem.cli.main.platform.config_file.update()

        # Load the Dev Env descriptors
        dem.cli.main.platform.load_dev_envs()

        # Run the CLI application
        dem.cli.main.typer_cli(prog_name=__command__)
    except LookupError as e:
        stderr.print("[red]" + str(e) + "[/]")
    except RegistryError as e:
        stderr.print("[red]" + str(e) + "\nUsing local tool images only![/]")
    except docker.errors.DockerException as e:
        stderr.print("[red]" + str(e) + "[/]")

        if "Permission denied" in str(e):
            stdout.print("\nHint: Is your user part of the docker group?")
        elif "invalid reference format" in str(e):
            stdout.print("\nHint: The input repository might not exist in the registry.")
        elif "400" in str(e):
            stdout.print("\nHint: The input parameters might not be valid.")
    except (ContainerEngineError, InternalError, ToolImageError, CatalogError) as e:
        stderr.print(f"[red]{str(e)}[/]")
    except DataStorageError as e:
        stderr.print("[red]" + str(e) + "[/]")
        if typer.confirm("Do you want to reset the file?"):
            if "config.json" in str(e):
                stdout.print("Restoring the original configuration file...")
                dem.cli.main.platform.config_file.restore()
            elif "dev_env.json" in str(e):
                stdout.print("Restoring the original Dev Env descriptor file...")
                dem.cli.main.platform.dev_env_json.restore()

# Call the main() when run as `python -m`
if __name__ == "__main__":
    main()



"""list-tools command implementation."""
# dem/cli/command/list_tools_cmd.py

from dem.core.platform import Platform
from dem.core.tool_images import ToolImage
from dem.cli.console import stdout, stderr
from rich.table import Table
import typer

def list_local_tools(platform: Platform) -> None:
    """ List the local tools.
    
        Args:
            platform -- the Platform

        Exceptions:
            typer.Abort -- if no local tool images are available
    """
    # by settings this to True, the update method won't try to update the registry tools
    platform.local_only = True

    if not platform.tool_images.all_tool_images:
        stdout.print("[yellow]No local tool images are available.[/]")
        raise typer.Abort()

    stdout.print(f"\n [italic]Local Tool Images[/]")
    # sorted will return a list of tuples, so we can iterate over the items
    for tool_image in sorted(platform.tool_images.all_tool_images.items()):
        # tool_image[0] is the name of the tool image and tool_image[1] is the ToolImage instance
        stdout.print(f"  {tool_image[0]}")

def update_tools_from_selected_regs(platform: Platform, selected_regs: list[str]) -> None:
    """ Update the tools from the selected registries only. 
    
        Args:
            platform -- the Platform
            selected_regs -- the selected registry names

        Exceptions:
            typer.Abort -- if an unknown registry is specified
    """
    # by settings this to True, the update method won't start automatically so we can provide
    # the selected registry names
    platform.disable_tool_update = True

    available_regs = set([reg["name"] for reg in platform.config_file.registries])
    selected_regs = set(selected_regs)

    if not selected_regs.issubset(available_regs):
        for unkown_reg in selected_regs - available_regs:
            stderr.print(f"[red]Error: Registry {unkown_reg} is not available![/]")
        raise typer.Abort()

    platform.tool_images.update(reg_selection=selected_regs)

def list_tools_from_regs(platform: Platform, table: Table) -> None:
    """ List the available tools from the registries.
    
        Args:
            platform -- the Platform
            table -- the rich Table instance to fill with the tool image data
    """
    table.add_column("Name")
    table.add_column("Available locally?")
    # sorted will return a list of tuples, so we can iterate over the items
    for tool_image in sorted(platform.tool_images.get_registry_ones().items()):
        # tool_image[0] is the name of the tool image and tool_image[1] is the ToolImage instance
        tool_image_name = tool_image[0]
        tool_image: ToolImage = tool_image[1]
        
        if tool_image.availability == ToolImage.LOCAL_AND_REGISTRY:
            available_locally = "[green]✔[/]"
        else:
            available_locally = ""

        table.add_row(tool_image_name, available_locally)

def list_tools_from_selected_regs(platform: Platform, specified_regs: list[str]) -> None:
    """ List the available tools from the selected registries only.
    
        Args:
            platform -- the Platform
            specified_regs -- the selected registry names
            
        Exceptions:
            typer.Abort -- if no tool images are available in the selected registries
    """
    update_tools_from_selected_regs(platform, specified_regs)

    if not platform.tool_images.get_registry_ones():
        stdout.print("[yellow]No tool images are available in the selected registries.[/]")
        raise typer.Abort()

    table = Table()
    list_tools_from_regs(platform, table)
    stdout.print(f"\n [italic]Available Tool Images from the selected registries[/]")
    stdout.print(table)

def list_tools_from_all_regs(platform: Platform) -> None:
    """ List the available tools from all registries.
    
        Args:
            platform -- the Platform
        
        Exceptions:
            typer.Abort -- if no tool images are available in the registries
    """
    if not platform.tool_images.get_registry_ones():
        stdout.print("[yellow]No tool images are available in the registries.[/]")
        raise typer.Abort()

    table = Table()
    list_tools_from_regs(platform, table)
    stdout.print(f"\n [italic]Available Tool Images from all registries[/]")
    stdout.print(table)

def execute(platform: Platform, reg: bool, selected_regs: list[str]) -> None:
    """ List the available tools.
        
        Args:
            platform -- the Platform
            reg -- the flag to list the tools from the registries
            selected_regs -- the selected registry names
        
        Exceptions:
            typer.Abort -- if no tool images are available either locally or in the registries or 
                           if an unknown registry is specified
    """
    if not reg:
        list_local_tools(platform)
    elif selected_regs:
        list_tools_from_selected_regs(platform, selected_regs)
    else:
        list_tools_from_all_regs(platform)


"""Repesents the Development Platform. The platform resources can be accessed through this interface.  
"""

import os
from typing import Any
from dem.core.core import Core
from dem.core.properties import __supported_dev_env_major_version__
from dem.core.exceptions import DataStorageError, PlatformError, ContainerEngineError
from dem.core.dev_env_catalog import DevEnvCatalogs
from dem.core.data_management import LocalDevEnvJSON
from dem.core.container_engine import ContainerEngine
from dem.core.registry import Registries
from dem.core.tool_images import ToolImages, ToolImage
from dem.core.dev_env import DevEnv
from dem.core.hosts import Hosts

class Platform(Core):
    """ Representation of the Development Platform:
        - The available tool images.
        - The available Development Environments.
        - External resources.
    """

    def _dev_env_json_version_check(self, dev_env_json_major_version: int) -> None:
        """ Check that the json file is supported.

            The version is stored as a string in the X.Y format.
            Raises an DataStorageError exception, if the version is not supported.
        """
        
        if dev_env_json_major_version != __supported_dev_env_major_version__:
            raise DataStorageError("The dev_env.json version v1.0 is not supported.")

    def __init__(self) -> None:
        """ Init the class."""
        self._dev_env_catalogs: DevEnvCatalogs | None = None
        self._tool_images = None
        self._container_engine = None
        self._registries = None
        self._hosts = None

        # Set this to true in the platform instance to work with the local tool images only
        self.local_only = False
        # Set this to true in the platform instance so when first accessing the `tool_images` 
        # instance variable, do not automatically update the tool images from the registries
        self.disable_tool_update = False

    def load_dev_envs(self) -> None:
        """ Load the Development Environments from the dev_env.json file.
        
            The Dev Envs will only contain the descriptors and not the ToolImage instances.
        """
        self.dev_env_json = LocalDevEnvJSON()
        self.dev_env_json.update()
        self.version = self.dev_env_json.deserialized["version"]
        self._dev_env_json_version_check(int(self.version.split('.', 1)[0]))
        self.local_dev_envs: list[DevEnv] = []
        for dev_env_descriptor in self.dev_env_json.deserialized["development_environments"]:
            self.local_dev_envs.append(DevEnv(descriptor=dev_env_descriptor))

    def assign_tool_image_instances_to_all_dev_envs(self) -> None:
        """ Assign the ToolImage instances to all Development Environments."""
        for dev_env in self.local_dev_envs:
            dev_env.assign_tool_image_instances(self.tool_images)

    @property
    def tool_images(self) -> ToolImages:
        """ The tool images.

            The ToolImages() gets instantiated only at the first access.
        """
        if self._tool_images is None:
            self._tool_images = ToolImages(self.container_engine, self.registries)
            if not self.disable_tool_update:
                self._tool_images.update(local_only=self.local_only)
        return self._tool_images
    
    @property
    def container_engine(self) -> ContainerEngine:
        """ The container engine.

            The ContainerEngine() gets instantiated only at the first access.
        """
        if self._container_engine is None:
            self._container_engine = ContainerEngine()

        return self._container_engine

    @property
    def registries(self) -> Registries:
        """ The registries.

            The Registries() gets instantiated only at the first access.
        """
        if self._registries is None:
            self._registries = Registries(self.container_engine)

        return self._registries

    @property
    def dev_env_catalogs(self) -> DevEnvCatalogs:
        """ The Development Environment Catalogs.

            The DevEnvCatalogs() gets instantiated only at the first access.
        """
        if self._dev_env_catalogs is None:
            self._dev_env_catalogs = DevEnvCatalogs()

        return self._dev_env_catalogs

    @property
    def hosts(self) -> Hosts:
        """ The hosts.
        
            The Hosts() gets instantiated only at the first access.
        """
        if self._hosts is None:
            self._hosts = Hosts()

        return self._hosts

    def get_deserialized(self) -> dict:
            """ Create the deserialized json. 
            
                Return the dev_env.json as a dict.
            """
            dev_env_json_deserialized: dict[str, Any] = {
                "version": self.version,
                "development_environments": [
                    dev_env.get_deserialized()
                    for dev_env in self.local_dev_envs
                ]
            }
            return dev_env_json_deserialized

    def get_dev_env_by_name(self, dev_env_name: str) -> DevEnv | None:
        """ Get the Development Environment by name.
        
            Args:
                dev_env_name -- name of the Development Environment to get

            Return with the instance representing the Development Environment. If the Development 
            Environment doesn't exist in the setup, return with None.
        """
        for dev_env in self.local_dev_envs:
            if dev_env.name == dev_env_name:
                return dev_env

    def install_dev_env(self, dev_env_to_install: DevEnv) -> None:
        """ Install the Dev Env by pulling the required images.
        
            Args:
                dev_env_to_install -- the Development Environment to install
        """
        # First check if the missing images are available in the registries, so DEM won't start to 
        # pull the images and then fail.
        for tool_image in dev_env_to_install.tool_images:
            if tool_image.availability == ToolImage.NOT_AVAILABLE:
                raise PlatformError(f"The {tool_image.name} image is not available.")

        for tool_image in dev_env_to_install.tool_images:
            if tool_image.availability == ToolImage.REGISTRY_ONLY:
                self.user_output.msg(f"\nPulling image {tool_image.name}", is_title=True)
                try:                
                    self.container_engine.pull(tool_image.name)
                except ContainerEngineError as e:
                    raise PlatformError(f"Dev Env install failed. --> {str(e)}")

        dev_env_to_install.is_installed = True
        self.flush_descriptors()

    def uninstall_dev_env(self, dev_env_to_uninstall: DevEnv) -> None:
        """ Uninstall the Dev Env by removing the images not required anymore.

            Args:
                dev_env_to_uninstall -- the Development Environment to uninstall

            Exceptions:
                PlatformError -- if the uninstall fails
        """
        all_required_tool_images = set()
        for dev_env in self.local_dev_envs:
            if (dev_env is not dev_env_to_uninstall) and dev_env.is_installed:
                for tool_image_descriptor in dev_env.tool_image_descriptors:
                    all_required_tool_images.add(tool_image_descriptor["image_name"] + ":" + tool_image_descriptor["image_version"])

        tool_images_to_remove = set()
        for tool_image_descriptor in dev_env_to_uninstall.tool_image_descriptors:
            tool_image_name = tool_image_descriptor["image_name"] + ":" + tool_image_descriptor["image_version"]
            if tool_image_name not in all_required_tool_images:
                tool_images_to_remove.add(tool_image_name)

        for tool_image_name in tool_images_to_remove:
            try:
                self.container_engine.remove(tool_image_name)
            except ContainerEngineError as e:
                raise PlatformError(f"Dev Env uninstall failed. --> {str(e)}")
            
        dev_env_to_uninstall.is_installed = False
        self.flush_descriptors()

    def flush_descriptors(self) -> None:
        """ Writes the deserialized json to the dev_env.json file."""
        # Get the up-to-date deserialized data.
        self.dev_env_json.deserialized = self.get_deserialized()
        self.dev_env_json.flush()

    def assign_dev_env(self, dev_env_to_assign: DevEnv, project_path: str) -> None:
        """ Assign the Development Environment to the project, by exporting the Dev Env's desriptor
            to the project's .axem directory.
        
            Args:
                dev_env_to_assign -- the Development Environment to assign
                project_path -- the path of the project
        """
        self.user_output.msg(f"\nAssigning the {dev_env_to_assign.name} Development Environment to the project at {project_path}")

        path: str = f"{project_path}/.axem"
        if not os.path.isdir(path):
            os.mkdir(path)

        path = f"{path}/dev_env_descriptor.json"
        if os.path.exists(path):
            self.user_output.get_confirm("[yellow]A Dev Env is already assigned to the project.[/]", 
                                         "Overwrite it?")

        dev_env_to_assign.export(path)

    def init_project(self, project_path: str) -> None:
        """ Init the project by saving the Dev Env's descriptor to the local Dev Env storage.

            Args:
                assigned_dev_env -- the Development Environment assigned to the project
        """
        descriptor_path = f"{project_path}/.axem/dev_env_descriptor.json"
        if not os.path.exists(descriptor_path):
            raise FileNotFoundError(f"The {descriptor_path} file does not exist.")

        assigned_dev_env = DevEnv(descriptor_path=descriptor_path)
        assigned_dev_env.assign_tool_image_instances(self.tool_images)
        existing_dev_env = self.get_dev_env_by_name(assigned_dev_env.name)
        if existing_dev_env is not None:
            self.user_output.get_confirm("[yellow]This project is already initialized.[/]", 
                                         "Overwrite it?")
            self.local_dev_envs.remove(existing_dev_env)

        self.local_dev_envs.append(assigned_dev_env)


"""Hosts."""
# dem/core/hosts.py

from dem.core.data_management import ConfigFile
from dem.core.core import Core

class Host():
    """ A Host. """
    def __init__(self, host_config: dict) -> None:
        """ Init the class with the host config.

            Args:
                host_config -- the host config
        """
        self.config: dict = host_config
        self.name: str = host_config["name"]
        self.address: str = host_config["address"]

class Hosts(Core):
    """ List of the available Hosts. """
    def __init__(self) -> None:
        """ Init the class with the host configurations.

            Args:
                config_file -- contains the host configurations
            """
        self.hosts: list[Host] = []
        for host_config in self.config_file.hosts:
            self._try_to_add_host(host_config)

    def _try_to_add_host(self, host_config: dict) -> bool:
        try:
            self.hosts.append(Host(host_config))
        except Exception as e:
            self.user_output.error(str(e))
            self.user_output.error("Error: Couldn't add this Host.")
            return False
        else:
            return True

    def add_host(self, host_config: dict) -> None:
        """ Add a new host.
        
            Args:
                catalog_config -- the new catalog to add
        """
        if self._try_to_add_host(host_config):
            self.config_file.hosts.append(host_config)
            self.config_file.flush()

    def list_host_configs(self) -> list[dict]:
        """ List the host configs. (As stored in the config file.)
        
            Return with the list of the available host configurations.
        """
        return self.config_file.hosts

    def delete_host(self, host_config: dict) -> None:
        """ Delete the host.
        
            Args:
                host_config -- config of the host to delete
        """
        for host in self.hosts.copy():
            if host.config == host_config:
                self.hosts.remove(host)

        self.config_file.hosts.remove(host_config)
        self.config_file.flush()


"""export CLI command implementation."""
# dem/cli/command/export_cmd.py


import os
from dem.core.platform import Platform
from dem.core.dev_env import DevEnv
from dem.cli.console import stderr

def export(dev_env: DevEnv, export_path: str):
    if os.path.isdir(export_path):
        export_path = f"{export_path}/{dev_env.name}"
    elif "" == export_path:
        export_path = dev_env.name
    
    if not export_path.endswith(".json"):
        export_path += ".json"

    dev_env.export(export_path)

def execute(platform: Platform, dev_env_name: str, export_path: str) -> None:


    dev_env = platform.get_dev_env_by_name(dev_env_name)
    if dev_env:
        try:
            export(dev_env, export_path)
        except FileNotFoundError:
            stderr.print("[red]Error: Invalid input path.[/]")
    else:
        stderr.print("[red]Error: The input Development Environment does not exist.[/]")


"""Image selector panel."""
# dem/cli/tui/panel/image_selector.py

from dem.cli.tui.renderable.menu import ToolImageMenu, DevEnvStatusPanel, CancelSaveMenu
from dem.cli.tui.printable_tool_image import PrintableToolImage
from rich.panel import Panel
from rich.layout import Layout
from rich.align import Align
from rich.live import Live
from rich.table import Table
from readchar import readkey, key

class NavigationHint(Panel):
    hint_text = """
- [bold]move cursor[/]: up and down arrows or j and k keys
- [bold]toggle selection[/]: space or enter
- [bold]jump between save/cancel and the selector[/]: tab
"""
    def __init__(self) -> None:
        super().__init__(self.hint_text, title="Navigation", expand=False)

class ErrorMessage(Table):
    def __init__(self, message: str) -> None:
        message = f"[red]{message}[/]"
        super().__init__(message, box=None, expand=False, show_edge=False)

class DevEnvSettingsWindow():
    def __init__(self, printable_tool_images: list[PrintableToolImage], 
                 already_selected_tool_images: list[str] = []) -> None:
        # Panel content
        self.dev_env_status_height = len(printable_tool_images)
        self.dev_env_status_width = 0
        for printable_tool_image in printable_tool_images:
            if len(printable_tool_image.name) > self.dev_env_status_width:
                self.dev_env_status_width = len(printable_tool_image.name)

        self.tool_image_menu = ToolImageMenu(printable_tool_images, already_selected_tool_images)
        self.dev_env_status_panel = DevEnvStatusPanel(already_selected_tool_images, 
                                                      self.dev_env_status_height, 
                                                      self.dev_env_status_width)
        self.cancel_save_menu = CancelSaveMenu()
        self.navigation_hint_panel = NavigationHint()
        self.error_message_panel = None

        self.build_layout()
        self.cancel_save_menu.remove_cursor()
        self.active_menu = self.tool_image_menu

    def build_layout(self) -> None:
        # Set the alignments
        aligned_tool_image_menu = Align(self.tool_image_menu, vertical="bottom", align="right")
        aligned_dev_env_status_panel = Align(self.dev_env_status_panel, vertical="bottom", align="left")
        aligned_cancel_save_menu = Align(self.cancel_save_menu, vertical="middle", align="center")
        aligned_navigation_hint_panel = Align(self.navigation_hint_panel, vertical="top", align="center")

        if self.error_message_panel:
            aligned_error_message_panel = Align(self.error_message_panel, vertical="middle", align="center")
        else:
            aligned_error_message_panel = Align(Table(box=None), vertical="middle", align="center")

        self.layout = Layout(name="root")

        self.layout.split_column(
            Layout(name="upper_half"),
            Layout(name="lower_half"),
        )
        self.layout["upper_half"].split_row(
            Layout(name="available"),
            Layout(name="selected")
        )

        self.layout["lower_half"].split_column(
            Layout(name="cancel_save", ratio=2),
            Layout(name="navigation_hint", ratio=6),
            Layout(name="error", ratio=2),
        )

        self.layout["available"].update(aligned_tool_image_menu)
        self.layout["selected"].update(aligned_dev_env_status_panel)
        self.layout["cancel_save"].update(aligned_cancel_save_menu)
        self.layout["navigation_hint"].update(aligned_navigation_hint_panel)
        self.layout["error"].update(aligned_error_message_panel)

    def _move_console(self) -> None:
        self.active_menu.remove_cursor()

        if self.active_menu is self.tool_image_menu:
            self.active_menu = self.cancel_save_menu
        else:
            self.active_menu = self.tool_image_menu

        self.active_menu.add_cursor()

    def _handle_user_input(self, input: str, live: Live) -> None:
        self.active_menu.handle_user_input(input)

        if (self.active_menu is self.tool_image_menu) and (input in [key.ENTER, key.SPACE]):
            self.dev_env_status_panel = DevEnvStatusPanel(self.tool_image_menu.get_selected_tool_images(),
                                                        self.dev_env_status_height,
                                                        self.dev_env_status_width)
            self.build_layout()
            live.update(self.layout)
        elif (self.active_menu is self.cancel_save_menu) and (input is key.ENTER) and \
            (not self.tool_image_menu.get_selected_tool_images()) and \
            ("save" in self.cancel_save_menu.get_selection()):
            self.error_message_panel = ErrorMessage("Error: No Tool Images selected.")
            self.cancel_save_menu.is_selected = False
            self.build_layout()
            live.update(self.layout)

    def wait_for_user(self) -> None:
        with Live(self.layout, refresh_per_second=8, screen=True) as live:
            while self.cancel_save_menu.is_selected is False:
                input = readkey()
                if input is key.TAB:
                    self._move_console()

                    if self.error_message_panel:
                        self.error_message_panel = None
                        self.build_layout()
                        live.update(self.layout)
                else:
                    self._handle_user_input(input, live)


"""rename CLI command implementation."""
# dem/cli/command/rename_cmd.py

from dem.core.platform import Platform
from dem.cli.console import stderr

def execute(platform: Platform, dev_env_name_to_rename: str, new_dev_env_name: str) -> None:


    dev_env_to_rename = platform.get_dev_env_by_name(dev_env_name_to_rename)

    if dev_env_to_rename is not None:
        dev_env_to_rename.name = new_dev_env_name
        platform.flush_descriptors()
    else:
        stderr.print("[red]Error: The input Development Environment does not exist.[/]")


"""info CLI command implementation."""
# dem/cli/command/info_cmd.py

from dem.core.tool_images import ToolImage
from dem.core.dev_env import DevEnv, DevEnv
from dem.cli.console import stdout, stderr
from dem.core.platform import Platform
from rich.table import Table
import typer

image_status_messages = {
    ToolImage.NOT_AVAILABLE: "[red]Error: not available![/]",
    ToolImage.LOCAL_ONLY: "Local",
    ToolImage.REGISTRY_ONLY: "Registry",
    ToolImage.LOCAL_AND_REGISTRY: "Local and Registry",
}

def print_local_dev_env_info(dev_env: DevEnv) -> None:
    """ Print information about the given local Development Environment.
    
        Args:
            dev_env -- the Development Environment to print information about
    """
    tool_info_table = Table()
    tool_info_table.add_column("Image")
    tool_info_table.add_column("Availability")

    for tool_image in sorted(dev_env.tool_images, key=lambda x: x.name):
        tool_info_table.add_row(tool_image.name,
                                image_status_messages[tool_image.availability])
    if dev_env.is_installed:
        installation_status = "[green]Installed[/]"
    else:
        installation_status = "Not installed"
    stdout.print(f"\n[bold]Development Environment: {dev_env.name}[/]\n")
    stdout.print(f"Installation state: {installation_status}\n")
    stdout.print(tool_info_table)

    if dev_env.is_installed and dev_env.get_tool_image_status() == DevEnv.Status.REINSTALL_NEEDED:
        stderr.print("\n[red]Error: Incomplete local install! The Dev Env must be reinstalled![/]")

    if dev_env.get_tool_image_status() == DevEnv.Status.UNAVAILABLE_IMAGE:
        stderr.print("\n[red]Error: Required image could not be found either locally or in the registry![/]")

def local_info(platform: Platform, dev_env_name: str) -> None:
    """ Gather and print information about the given local Development Environment.
    
        Args:
            platform -- the platform
            dev_env_name -- the name of the Development Environment to print information about

        Raises:
            typer.Abort -- if the Development Environment is unknown
    """
    platform.assign_tool_image_instances_to_all_dev_envs()

    dev_env = platform.get_dev_env_by_name(dev_env_name)

    if dev_env is None:
        stderr.print(f"[red]Error: Unknown Development Environment: {dev_env_name}[/]\n")
        raise typer.Abort()

    print_local_dev_env_info(dev_env)

def print_cat_dev_env_info(dev_env: DevEnv, cat_name: str) -> None:
    """ Print information about the given catalog Development Environment.
    
        Args:
            dev_env -- the Development Environment to print information about
            cat_name -- the name of the catalog the Development Environment belongs to
    """
    tool_info_table = Table()
    tool_info_table.add_column("Image")
    tool_info_table.add_column("Availability")

    for tool_image in sorted(dev_env.tool_images, key=lambda x: x.name):
        tool_info_table.add_row(tool_image.name,
                                image_status_messages[tool_image.availability])
    stdout.print(f"\n[bold]Development Environment: {dev_env.name}[/]\n")
    stdout.print(f"Catalog: {cat_name}\n")
    stdout.print(tool_info_table)

    if dev_env.get_tool_image_status() == DevEnv.Status.UNAVAILABLE_IMAGE:
        stderr.print("\n[red]Error: Required image could not be found in the registry![/]")
    
def cat_dev_env_info(platform: Platform, dev_env_name: str, selected_cats: list[str]) -> None:
    """ Gather and print information about the given catalog Development Environment.
    
        Args:
            platform -- the platform
            dev_env_name -- the name of the Development Environment to print information about
            selected_cats -- the selected catalog names, empty list means all catalogs
    """
    for catalog in platform.dev_env_catalogs.catalogs:
        if catalog.name in selected_cats or not selected_cats:
            catalog.request_dev_envs()
            dev_env = catalog.get_dev_env_by_name(dev_env_name)
            if dev_env:
                dev_env.assign_tool_image_instances(platform.tool_images)
                print_cat_dev_env_info(dev_env, catalog.name)
                break
    else:
        stderr.print(f"[red]Error: Unknown Development Environment: {dev_env_name}[/]\n")

def selected_cats_info(platform: Platform, dev_env_name: str, selected_cats: list[str]) -> None:
    """ Print information about the given Development Environment from the selected catalogs.
    
        Args:
            platform -- the platform
            dev_env_name -- the name of the Development Environment to print information about
            selected_cats -- the selected catalog names
            
        Raises:
            typer.Abort -- if the selected catalog is unknown
    """
    available_cats = set([cat.name for cat in platform.dev_env_catalogs.catalogs])
    selected_cats = set(selected_cats)

    if not selected_cats.issubset(available_cats):
        stderr.print(f"[red]Error: Unknown catalog(s): {', '.join(selected_cats - available_cats)}[/]\n")
        raise typer.Abort()

    cat_dev_env_info(platform, dev_env_name, selected_cats)

def all_cats_info(platform: Platform, dev_env_name: str) -> None:
    """ Print information about the given Development Environment from all catalogs.
    
        Args:
            platform -- the platform
            dev_env_name -- the name of the Development Environment to print information about
    """
    cat_dev_env_info(platform, dev_env_name, [])

def execute(platform: Platform, arg_dev_env_name: str, cat: bool, selected_cats: list[str]) -> None:
    """ Print information about the given Development Environment.
    
        Args:
            platform -- the platform
            arg_dev_env_name -- the name of the Development Environment to print information about
            cat -- the flag to print information about the Development Environment from the catalogs
            selected_cats -- the selected catalog names

        Raises:
            typer.Abort -- if the Development Environment is unknown or if the selected catalog is 
                           unknown
    """
    if not cat:
        local_info(platform, arg_dev_env_name)
    elif selected_cats:
        selected_cats_info(platform, arg_dev_env_name, selected_cats)
    else:
        all_cats_info(platform, arg_dev_env_name)


"""CLI command implementation for listing the available registries."""
# dem/cli/command/list_reg_cmd.py

from dem.core.platform import Platform
from dem.cli.console import stdout
from rich.table import Table

def execute(platform: Platform) -> None:
    """ List available registries."""
    registry = None
    table = Table()
    table.add_column("name")
    table.add_column("url")

    for registry in platform.registries.list_registry_configs():
        table.add_row(registry["name"], registry["url"])
    
    if registry is None:
        stdout.print("[yellow]No available registries![/]")
    else:
        stdout.print(table)